import sys
import os
sys.path.insert(0,os.path.dirname(sys.executable)+'/../lib/python2.7/site-packages/') #ignore other PYTHONPATH

import logging
import docopt
import schema
import brilws
import yaml
import base64
from sqlalchemy import *
if sys.version_info[0] == 2:
    from ConfigParser import SafeConfigParser
else:
    from configparser import SafeConfigParser
from brilws import api
import pandas as pd

log = logging.getLogger('brilschema')
logformatter = logging.Formatter('%(levelname)s %(message)s')
ch = logging.StreamHandler()
ch.setLevel(logging.WARNING)
#fh = logging.FileHandler('/tmp/brilschema.log')
ch.setFormatter(logformatter)
#fh.setFormatter(logformatter)
log.addHandler(ch)
#log.addHandler(fh)


def brilschema_main(progname=sys.argv[0]):
    docstr='''


    usage:
      brilschema (-h|--help) 
      brilschema --version
      brilschema --checkforupdate
      brilschema [--debug|--nowarning] <command> [<args>...]

    commands:
      create      generate table creat/drop sql
      loadmap     load map data  
      loaddata    load data
      loadresult  load result

    See 'brilschema <command> --help' for more information on a specific command.

    '''

    args = {}
    argv = sys.argv[1:]
    args = docopt.docopt(docstr,argv,help=True,version=brilws.__version__,options_first=True)
    
    if '--debug' in sys.argv:
       log.setLevel(logging.DEBUG)
       ch.setLevel(logging.DEBUG)
    if args['--version'] : 
        print(brilws.__version__)
    log.debug('global arguments: %s',args)
    cmmdargv = [args['<command>']] + args['<args>']

    log.debug('command arguments: %s',cmmdargv)
    parseresult = {}

    try:
      if args['<command>'] == 'create':
         from . import brilschema_create
         parseresult = docopt.docopt(brilschema_create.__doc__,argv=cmmdargv)
         parseresult = brilschema_create.validate(parseresult)
         columntypemap={}
         schemadatadef={}
         infile = parseresult['-i']
         infilenamebase = os.path.basename(infile.name).split('.')[0]         
         suffix = parseresult['--suffix']             
         writeraccount = ''
         if parseresult['-f']=='oracle':
              columntypemap = api.oracletypemap
              if '--writer' not in parseresult:                  
                  writeraccount = parseresult['--writer']
         else:
              columntypemap = api.sqlitetypemap
         schemadatadef = yaml.load(parseresult['-i'])
         print ('Creating drop sql file for: %s'%infilenamebase)
         api.drop_tables_sql(infilenamebase,schemadatadef,suffix=suffix,dbflavor=parseresult['-f'])
         print ('Creating create sql file for: %s'%infilenamebase)
         api.create_tables_sql(infilenamebase,schemadatadef,suffix=suffix,dbflavor=parseresult['-f'],writeraccount=writeraccount)
         print ('Done')

      elif args['<command>'] == 'loadmap':
         from . import brilschema_loadmap         
         parseresult = docopt.docopt(brilschema_loadmap.__doc__,argv=cmmdargv)
         parseresult = brilschema_loadmap.validate(parseresult)
         inengine = os.path.expanduser(parseresult['-i'])
         outengine = os.path.expanduser(parseresult['-o'])
         iniparser = SafeConfigParser()
         if parseresult['--name'] == 'datatable':
             d = api.DatatableMap()
         elif parseresult['--name'] == 'amodetag':
             d = api.AmodetagMap()
         elif parseresult['--name'] == 'beamstatus':
             d = api.BeamStatusMap()
             result = d.from_sourcedb(inengine)
         elif parseresult['--name'] == 'trgbit':
             d = api.TrgBitMap()
         elif parseresult['--name'] == 'hltconfig':
             d = api.HltConfigMap()
         elif parseresult['--name'] == 'l1seed':
             d = api.L1SeedMap()
         elif parseresult['--name'] == 'hltpath':
             d = api.HLTPathMap()
         elif parseresult['--name'] == 'trghltseed':
             d = api.TrgHltSeedMap()
         elif parseresult['--name'] == 'dataset':
             d = api.DatasetMap()
         elif parseresult['--name'] == 'hltstreamdataset':
             d = api.HLTStreamDatasetMap()
         if not os.path.isfile(inengine):
             sourceurl = inengine
             if sourceurl.find('oracle') != -1:
                 iniparser.read(os.path.join(parseresult['-p'],'readdb2.ini'))
                 sourcepasswd = parser.get(outengine,'pwd')
                 sourceschema = parser.get(outengine,'schema')
                 idx = sourceengine.find('@')
                 code = base64.b64decode(sourcepasswd).decode('UTF-8')
                 sourceurl = sourceengine[:idx]+':'+code+sourceengine[idx:]
             #inengine = create_engine(soureceurl)
             if sys.version_info > (3,):
                 inengine = create_engine(sourceurl,max_identifier_length=128)
             else:
                 inengine = create_engine(sourceurl)
         result = d.from_sourcedb(inengine)
         print (result)
         if os.path.isfile(outengine):
             d.to_csv(outengine,result)
         else:
             desturl = outengine
             if desturl.find('oracle') != -1:
                 iniparser.read(os.path.join(parseresult['-p'],'authentication.ini'))
                 destpasswd = parser.get(outengine,'pwd')
                 idx = outengine.find('@')
                 destcode =  base64.b64decode(destpasswd).decode('UTF-8')
                 desturl = outengine[:idx]+':'+destcode+outengine[idx:]             
             #outengine = create_engine(desturl)
             if sys.version_info > (3,):
                 outengine = create_engine(desturl,max_identifier_length=128)
             else:
                 outengine = create_engine(desturl)
             d.to_brildb(outengine,result)
      elif args['<command>'] == 'loaddata':
         from . import brilschema_loaddata         
         parseresult = docopt.docopt(brilschema_loadmap.__doc__,argv=cmmdargv)
         parseresult = brilschema_loaddata.validate(parseresult)
         incsv = os.path.expanduser(parseresult['-i'])
         c = api.get_filepath_or_buffer(incsv)
         if os.path.isfile(c):
             pass
         else:
             pass
         outengine = os.path.expanduser(parseresult['-o'])
         iniparser = SafeConfigParser()
         if parseresult['--name'] == 'fillinfo':
             d = api.FillInfo()
             if parseresult['--lumidb']:
                 d.from_lumidb(lumidbengine,fillnum)
                 print (d)
      elif args['<command>'] == 'loadresult':
         from . import brilschema_loadresult
         iniparser = SafeConfigParser()
         parseresult = docopt.docopt(brilschema_loadresult.__doc__,argv=cmmdargv)
         parseresult = brilschema_loadresult.validate(parseresult)
         incsv = os.path.expanduser(parseresult['-i'])
         runlsselect = api.parsecmsselectJSON(runlsselect)
         lumidb  = api.get_filepath_or_buffer(parseresult['--lumidb'])
         if os.path.isfile(lumidb):
             print (lumidb)
         else:
             lumidburl = lumidb
             if lumidburl.find('oracle') != -1:
                 iniparser.read(os.path.join(parseresult['-p'],'authentication.ini'))
                 lumidbpasswd = parser.get(lumidburl,'pwd')
                 idx = lumidburl.find('@')
                 lumicode = base64.b64decode(lumidbpasswd).decode('UTF-8')
                 lumidburl = lumidburl[:idx]+':'+lumicode+lumidburl[idx:]             
             if sys.version_info > (3,):
                 lumidbengine = create_engine(lumidburl,max_identifier_length=128)
             else:
                 lumidbengine = create_engine(lumidburl)
         datasourcename = parseresult['--name'].lower()
         d = api.LumiResult(datasourcename)
         for run,lss in runlsselect:
             print (run,lss)
         #d.from_lumidb(lumidbengine,)
         
      else:
         exit("%r is not a brilschema command. See 'brilschema --help'."%args['<command>'])
    except docopt.DocoptExit:
      raise docopt.DocoptExit('Error: incorrect input format for '+args['<command>'])
    except schema.SchemaError as e:
      exit(e)

    if '--debug' not in parseresult:
       if '--nowarning' not in parseresult:
          log.setLevel(logging.ERROR)
          ch.setLevel(logging.ERROR)
    else:
        log.setLevel(logging.DEBUG)
        ch.setLevel(logging.DEBUG)
    log.debug('create arguments: %s',parseresult)

    
if __name__ == '__main__':
    brilschema_main()



"""    
Usage:
  briltag listiov [options]


Options:
      -h --help                Show this screen.
      -c CONNECT               Service name [default: offline]
      -p AUTHPATH              Authentication file
      --name TAGNAME           IOV tag name 
      --type DATASOURCE        Data source type. bcm1f,bcm1fsi,bcm1futca,plt,pxl,hfoc,hfet
      --applyto APPLYTO        Type of correction applied. lumi
      --isdefault              Show only default tags [default: False]

"""

from docopt import docopt
from schema import Schema
from brilws.cli import clicommonargs

def validate(optdict):
    myvalidables = ['-c','-p','--name','--applyto','--type',str]
    argdict = dict((k,v) for k,v in clicommonargs.argvalidators.items() if k in myvalidables)
    s = Schema(argdict)
    result = s.validate(optdict)
    return result

if __name__ == '__main__':
    print (docopt(__doc__,options_first=True))





import re
class RegexValidator(object):
    def __init__(self, pattern, statement=None):
        self.pattern = re.compile(pattern)
        self.statement = statement
        if not self.statement:
           self.statement = "must match pattern %s" % self.pattern
    def __call__(self, string):
        match = self.pattern.search(string)
        if not match:
            raise ValueError(self.statement)
        return string



import numpy as np

#dec_prec = 9
#dec_prec_value = 10 ** (-1 * dec_prec)

def bxintensity(x):    
    return '{:d} {:.4e} {:.4e}'.format(int(x[0]),x[1],x[2])

def bitprescFormatter(x):
    bitname = x[0]
    prescval = x[1]
    return '%s/%.2f'%(bitname,prescval)

class Formatter:
    def __init__(self, formatstr, prec):
        self.formatstr = formatstr
        self.prec = prec
        self.dec_prec_value = None
        if self.formatstr=='f':
            self.dec_prec_value = 10 ** (-1 * self.prec)

    def lumi(self,x):
        if self.formatstr=='e':
            return self.lumiE(x)
        if abs(x)<self.dec_prec_value:
            return '0'
        return '{:.{}f}'.format(float(x),self.prec)

    def bxlumi(self,x):
        if self.formatstr=='e':
            return self.bxlumiE(x)
        bxid = int(x[0])
        bxid_str = '{:d}'.format(bxid)
        vd = x[1]
        if abs(vd)<self.dec_prec_value:    
            return ' '.join([bxid_str,'0','0'])
        vd_str = '{:.{}f}'.format(vd,self.prec)
        vr = x[2]
        if abs(vr)<self.dec_prec_value:
            return ' '.join([bxid_str,vd_str,'0'])
        vr_str = '{:.{}f}'.format(vr,self.prec)
        return ' '.join([bxid_str,vd_str,vr_str])

    def lumiE(self,x):
        v_str = '{:.{}}e'.format(float(x),self.prec)
        return v_str

    def bxlumiE(self,x):
        bxid = int(x[0])
        vd = x[1]
        vr = x[2]
        bxid_str = '{:d}'.format(bxid)
        vd_str = '{:.{}}e'.format(vd,self.prec)
        vr_str = '{:.{}}e'.format(vr,self.prec)
        return ' '.join([bxid_str,vd_str,vr_str])
    
    @property
    def lumiunit_to_scalefactor(self):
        return {    
        '/kb':1.0e-09,
        '1e21/cm2':1.0e-09,
        'hz/kb':23.31*1.0e-09,
        '1e21/cm2s':23.31*1.0e-09,        

        '/b':1.0e-06,
        '1e24/cm2':1.0e-06,
        'hz/b':23.31*1.0e-06,
        '1e24/cm2s':23.31*1.0e-06,
    
        '/mb':1.0e-03,
        '1e27/cm2':1.0e-03,
        'hz/mb':23.31*1.0e-03,
        '1e27/cm2s':23.31*1.0e-03,
        
        '/ub':1.0,
        '1e30/cm2':1.0,
        'hz/ub':23.31,
        '1e30/cm2s':23.31,
        
        '/nb':1.0e3,
        '1e33/cm2':1.0e3,
        'hz/nb':23.31*1.0e3,
        '1e33/cm2s':23.31*1.0e3,
        
        '/pb':1.0e6,
        '1e36/cm2':1.0e6,
        'hz/pb':23.31*1.0e6,
        '1e36/cm2s':23.31*1.0e6,
        
        '/fb':1.0e9,
        '1e39/cm2':1.0e9,
        'hz/fb':23.31*1.0e9,
        '1e39/cm2s':23.31*1.0e9,
        
        '/ab':1.0e12,
        '1e42/cm2':1.0e12,
        'hz/ab':23.31*1.0e12,
        '1e42/cm2s':23.31*1.0e12,
         }

    def lumiunit(self,x,unitstr):
        if x.find('(/ub)')==-1: 
            return x
        return x.replace('(/ub)','(%s)'%unitstr)
    
#def lumiunit(x,scalefactor):
#    '''
#    return lumiunit according to scalefactor 
#    if x does not contain string /ub, do nothing
#    special if scalefactor is about 23, return hz/ub
#    else:
#        if scalefactor>=1.0e-03 and scalefactor<1, replace /ub with /mb
#        if scalefactor>=1 and scalefactor<1e3, keep /ub
#        if scalefactor>=1e3 and scalefactor<1e6, replace /ub with /nb
#        if scalefactor>=1e6 and scalefactor<1e9, replace /ub with /pb
#        if scalefactor>=1e9 and scalefactor<1e12, replace /ub with /fb
#        if scalefactor>=1e12 and scalefactor<1e15, replace /ub with /ab
#    '''
    
#    if x.find('(/ub)')==-1: return x    
#    if int(scalefactor) == 23: return x.replace('(/ub)','(hz/ub)')
#    if (scalefactor>=1.0e-09) and (scalefactor<1.0e-06): return x.replace('(/ub)','(/kb)')
#    if (scalefactor>=1.0e-06) and (scalefactor<1.0e-03): return x.replace('(/ub)','(/b)')
#    if (scalefactor>=1.0e-03) and (scalefactor<1.0): return x.replace('(/ub)','(/mb)')
#    if (scalefactor>=1.0) and (scalefactor<1.0e3): return x
#    if (scalefactor>=1.0e3) and (scalefactor<1.0e6): return x.replace('(/ub)','(/nb)')
#    if (scalefactor>=1.0e6) and (scalefactor<1.0e9): return x.replace('(/ub)','(/pb)')
#    if (scalefactor>=1.0e9) and (scalefactor<1.0e12): return x.replace('(/ub)','(/fb)')
#    if (scalefactor>=1.0e12) and (scalefactor<1.0e15): return x.replace('(/ub)','(/ab)')
#    return x

    def guesslumiUnit(self,inverseubval):
        '''
        input:
        float value in 1/ub
        output:
        printable value (value(float),unit(str)) unit in [1/kb,1/b,1/mb,1/ub,1/nb,1/pb,1/fb]
        '''
        denomitor = 1.0
        if inverseubval>=1.0e-09 and inverseubval<1.0e-06:
            denomitor=1.0e-09
            unitstring='/kb'
            return (float(inverseubval)/float(denomitor),unitstring)
        if inverseubval>=1.0e-06 and inverseubval<1.0e-03:
            denomitor=1.0e-06
            unitstring='/b'
            return (float(inverseubval)/float(denomitor),unitstring)
        if inverseubval>=1.0e-03 and inverseubval<1.0:
            denomitor=1.0e-03
            unitstring='/mb'
            return (float(inverseubval)/float(denomitor),unitstring)
        if inverseubval>=1.0 and inverseubval<1.0e3:
            unitstring='/ub'
            return (inverseubval,unitstring)
        if inverseubval>=1.0e3 and inverseubval<1.0e06:
            denomitor=1.0e3
            unitstring='/nb'
            return (float(inverseubval)/float(denomitor),unitstring)
        if inverseubval>=1.0e6 and inverseubval<1.0e9:
            denomitor=1.0e6
            unitstring='/pb'
            return (float(inverseubval)/float(denomitor),unitstring)
        if inverseubval>=1.0e9 and inverseubval<1.0e12:
            denomitor=1.0e9
            unitstring='/fb'
            return (float(inverseubval)/float(denomitor),unitstring)
        if inverseubval>=1.0e12 and inverseubval<1.0e15:
            denomitor=1.0e12
            unitstring='/ab'
            return (float(inverseubval)/float(denomitor),unitstring)
        return (float(inverseubval),denomitor,'/ub')






import ast
import numpy as np
import collections
import copy
from numpy.polynomial import polynomial as P

class FunctionRoot(object):
    def __init__(self,*args):
        self.root = []
        for a in args:
            self.root.append(a)
            
class CorrectionFunction(object):
    def __init__(self,fname,fparams):
        self.fname = fname
        self.fparams = fparams
    def name(self):
        return self.fname
    def params(self):
        return self.fparams

def applyCorrectionStr(funcstrs , paramstrs, funcroot ):
        funcs_list = [ funcstrs.strip() ]
        params_list = [ paramstrs.strip() ]
        if funcstrs[0]=='[':
            funcs_list = ast.literal_eval(funcstrs)
        if paramstrs[0]=='[':
            params_list = ast.literal_eval(paramstrs)
        funcs = []       
        for f,p in zip(funcs_list,params_list):
            paramdict = p
            if isinstance(p,str):
                paramdict = ast.literal_eval(paramdict)
            funcs.append( CorrectionFunction(f, paramdict) )
        return applyCorrection( funcs, funcroot )                        
        
def applyCorrection(funcs, funcroot):
        '''
        input:
            funcroot:  FunctionRoot object (bxoravglumi,avglumi,ncollidingbx)
            funcs:  list of CorrectionFunction objects
        output: 
            (totlumi,bxlumi|None)
        '''
        fr = copy.deepcopy(funcroot)
        toproot = fr.root[0]
        withBX = True
        totlumi = fr.root[1]
        bxlumi = None
        if not isinstance(toproot,collections.Iterable) :
            withBX = False
        else:
            bxlumi = toproot

        for f in funcs:
            totlumi = FunctionCaller( f, FunctionRoot(totlumi,totlumi,fr.root[2]) )
            fr.root[0] = totlumi
            fr.root[1] = totlumi
            if withBX:
                bxlumi = FunctionCaller( f, FunctionRoot(bxlumi,totlumi,fr.root[2]) )
                fr.root[0] = bxlumi
        return (totlumi,bxlumi)
    
class FunctionFactory(object):
    def hcaldrift(self,ivalue,rawavglumi,intglumi,nbx,a0=1.0,a1=0.0,a2=0.0,drift=0.0,c1=0.0,afterglowthresholds=''):
        '''
        this is a pain . just dump here. make it usable again only if later maybe used.        
        '''
        avglumi = 0.
        if c1 and nbx>0:
            avglumi = c1*rawavglumi/nbx
        driftterm = 1.0
        if drift and intglumi: driftterm=1.0+drift*intglumi
        resultfactor = a0 * driftterm * self.inversepoly1d(avglumi,np.array([a2,a1,1])) 
        result = ivalue*resultfactor
        if afterglowthresholds:
            result = self.afterglow(result,nbx,afterglowthresholds)
        return result
    
    def poly1d(self,functionroot,kwds):
        '''
        root0: avglumi or bxlumi        
        '''
        ivalue = functionroot.root[0]
        ncollidingbx =  functionroot.root[2]
        coefsStr = kwds['coefs']
        coefs = np.fromstring(coefsStr, dtype=np.float, sep=',')
        if not isinstance(ivalue,collections.Iterable) : #is totallumi, in this case, need to use term totallumi/nbx
            with np.errstate(divide='ignore',invalid='ignore'):
                ivalue = np.true_divide(ivalue,ncollidingbx)
                if ivalue == np.inf:
                    ivalue = 0
                ivalue = np.nan_to_num(ivalue)
        if len(coefs)>1:
            coefs = coefs[::-1] #reverse the order because polyval coefs order is reverse of np.poly1d
        if isinstance(ivalue,collections.Iterable) :
            return P.polyval(ivalue,coefs)
        else:
            return ncollidingbx*P.polyval(ivalue,coefs)           
            
    def poly2dlL(self,functionroot,kwds):
        '''
        poly2d of l=bxlumi, L=totallumi
        c00 + c10*x + c01*y (n=1)
        c00 + c10*x + c01*y + c20*x**2 + c02*y**2 + c11*x*y (n=2)
        if l is scalar, scale l to y/NBX, then the final result is NBX*singleresult
        otherwise, make l,L  the same shape
        '''        
        l = functionroot.root[0]  #per bunch lumi or totallumi
        L = functionroot.root[1] #total lumi
        ncollidingbx = functionroot.root[2] 
        totlumi = 0
        bxlumi = 0

        if isinstance(l,collections.Iterable):
            bxlumi = np.array(l)
            totlumi = np.full_like(bxlumi, L)
        else:
            if l!=L:
                raise ValueError('l and L are of different value ')
            with np.errstate(divide='ignore',invalid='ignore'):
                bxlumi = np.true_divide(L,ncollidingbx)
                if bxlumi == np.inf:
                    bxlumi = 0
                bxlumi = np.nan_to_num(bxlumi)                    
            totlumi = L

       
        coefsStr = kwds['coefs']

        coefs = np.array(ast.literal_eval(coefsStr), dtype=np.float)
        if isinstance(l,collections.Iterable):
            return P.polyval2d(bxlumi,totlumi,coefs)
        else:
            return ncollidingbx*P.polyval2d(bxlumi,totlumi,coefs)

    def inversepoly1d(self,functionroot,kwds):
        return 1./self.poly1d(functionroot,kwds)
    
    def afterglow(self,functionroot,kwds):
        ivalue = functionroot.root[0]
        ncollidingbx = functionroot.root[2]
        afterglowthresholds = kwds['afterglowthresholds']
        afterglow = 1.

        if afterglowthresholds[-1]!=',':
            afterglowthresholds+=','
        afterglowthresholds = np.array(ast.literal_eval(afterglowthresholds))
        for (bxthreshold,c) in afterglowthresholds:
            if ncollidingbx >= bxthreshold:
                afterglow = c

        return ivalue*afterglow 

    def poly1dWafterglow( self,functionroot,kwds ):
        poly1dresult = self.poly1d(functionroot,kwds)
        fr = FunctionRoot(poly1dresult,functionroot.root[1],functionroot.root[2])
        result = self.afterglow(fr,kwds)
        return result

    def afterglowWpoly2dlL(self,functionroot,kwds):
        '''
        afterglow*poly2dlL
        '''
        l = functionroot.root[0]
        L = functionroot.root[1]
        afterglowl = self.afterglow( FunctionRoot(l,functionroot.root[1],functionroot.root[2]), kwds )
        afterglowL = self.afterglow( FunctionRoot(L,functionroot.root[1],functionroot.root[2]), kwds )
        fr = FunctionRoot(afterglowl,afterglowL,functionroot.root[2])
        result = self.poly2dlL(fr,kwds)
        return result

    def poly2dlLWafterglow(self,functionroot,kwds):
        '''
        poly2dlL*afterglow
        '''
        polyresult = self.poly2dlL(functionroot,kwds)
        fr = FunctionRoot(polyresult,functionroot.root[1],functionroot.root[2])
        result = self.afterglow(fr,kwds)
        return result
    
def FunctionCaller(func,functionroot):
    fac = FunctionFactory()
    try:
        myfunc = getattr(fac,func.fname,None)
    except AttributeError:
        print ('[ERROR] unknown correction function '+ func.fname)
        raise
    if callable(myfunc):
        return myfunc(functionroot, func.fparams)
    else:
        raise ValueError('uncallable function '+func.fname)
     
if __name__=='__main__':
    # test CorrectionFunction class
    f = CorrectionFunction('poly1d',{'coefs':'2.,1.,0.'})
    print (f.name(), f.params())
    nbx = 3
    ivalue =  np.array([2.,2.,2.])
    funcroot = FunctionRoot( ivalue, None, nbx)
    
    funcs = [ CorrectionFunction('poly1d',{'coefs':'2.,1.,0.'}) ]
    result = applyCorrection(funcs,funcroot)
    print (result)

    ivalue = 6
    fr = FunctionRoot( ivalue,ivalue,nbx)
    funcs = [ CorrectionFunction('poly1d',{'coefs':'2.,1.,0.'}) ]
    result =  applyCorrection(funcs,fr)
    print ('result  total ',result)

    funcs =  [ CorrectionFunction('afterglow',{'afterglowthresholds':'(1,2),'}) ]
    result = applyCorrection(funcs,fr)
    print ('afterglow result ',result)

    fr =  FunctionRoot([1,2,3],2,nbx)
    funcs =  [ CorrectionFunction('poly2dlL',{'coefs':'[[3,2,1],[-10,4,0],[2,0,0]]'}) ]
    result = applyCorrection(funcs,fr)
    print ('poly2dlL result ',result)

    fr = FunctionRoot(3,3,nbx)
    funcs =  [ CorrectionFunction('poly2dlL',{'coefs':'[[3,2,1],[-10,4,0],[2,0,0]]'}) ]
    result = applyCorrection(funcs,fr)
    print ('poly2dlL result ',result)
    
    result = applyCorrectionStr('poly2dlL' , "{'coefs':'[[3,2,1],[-10,4,0],[2,0,0]]'}", fr)
    print ('poly2dlL result from string ',result)

    result = applyCorrectionStr("['poly2dlL','afterglow']" , "[{'coefs':'[[3,2,1],[-10,4,0],[2,0,0]]'},{'afterglowthresholds':'(1,2),'}]", fr)
    print ('poly2dlL result from string ',result)



"""
Usage:
  brilschema create [options] 

options:
  -h, --help       Show this screen
  -i INPUTFILE     Input schema definition file
  -f DBFLAVOR      Database flavor [default: sqlite]
  --writer WRITER  DB Account with write permission
  --suffix SUFFIX  Table suffix 

"""

dbflavors=['oracle','sqlite']

import os
from docopt import docopt
from schema import Schema, And, Or, Use
from brilws.cli import clicommonargs

def validate(optdict):
    result={}
    schema = Schema({ 
     '-i': Use(open, error='-i INPUTFILE should be readable'),
     '-f': And(str,lambda s: s.lower() in dbflavors, error='-f must be in '+str(dbflavors)),
     '--writer': Or(None,Use(str.upper)),
     '--suffix': Or(None,Use(str.upper)),
     str:object # catch all
    })
    result=schema.validate(optdict)
    return result

if __name__ == '__main__':
    print (docopt(__doc__))



import tables as tb
import numpy as np
import pandas as pd
import os,sys
import logging
from time import time
from . import params

log = logging.getLogger('brilws')

def _is_element_tuple(me):
    '''
    Check if element of me is tuple 
    Input: collection
    '''
    return np.all([isinstance(x,tuple) for x in me])

def _is_subset(me,target):
    '''
    Check me collection is a subset of the target collection
    '''
    return np.in1d(me,target).all()

def _open_validfile(filename, tables):
    '''
    Check if the file has all the required table    
    Input: f: filename, tables: required table names
    Output: filehandle if input is valid, None if input is invalid 
    '''    

    f = None
    try:
        f = tb.open_file(filename)
    except tb.exceptions.HDF5ExtError:
        log.debug('failed to open %s, skip'%filename)
        return None
    nodes = list(f.root._v_children.keys())
    if not len(nodes):
        f.close()
        return None    
    if not _is_subset(tables,nodes):
        f.close()
        return None
    tablelist = [f.get_node('/'+x) for x in tables]
    filledtables = [x for x in tablelist if x.nrows>0]
    if len(tablelist) != len(filledtables):
        return None
    return f

def open_validfiles(filenames,lumitype):
    '''
    Input: [filename], lumitype
    Output: [filehandle]
    '''    
    mytype = lumitype.upper()
    requiredtables = ['tcds','beam']
    if mytype in params._lumitypeToh5tablename.keys():
        requiredtables.append(params._lumitypeToh5tablename[mytype])
    fhs = [_open_validfile(f,requiredtables) for f in filenames]
    return [x for x in fhs if x is not None]

def _build_preselectcondition( runmin=None,runmax=None,fillmin=None,tssecmin=None,tssecmax=None,fillmax=None, runlsselect=None ):
        if runlsselect is not None:        
            if runmin==runmax==fillmin==tssecmin==tssecmax==fillmax==None:
                runmin = min(runlsselect.index.tolist())
                runmax = max(runlsselect.index.tolist())       
        conditions = []
        if runmin and runmax and runmin==runmax:
            conditions.append('(runnum == %d)'%runmin)   
        elif runmin: 
            conditions.append('(runnum >= %d)'%runmin)
        elif runmax: 
            conditions.append('(runnum <= %d)'%runmax)
        if fillmin and fillmax and fillmin==fillmax:
            conditions.append('(fillnum == %d)'%fillmin)
        elif fillmin:
            conditions.append('(fillnum >= %d)'%fillmin)
        elif fillmax: 
            conditions.append('(fillnum <= %d)'%fillmax)
        if tssecmin and tssecmax and tssecmin==tssecmax:
            conditions.append('(tssec == %d)'%tssecmin)
        elif tssecmin: 
            conditions.append('(tssec >= %d)'%tssecmin)
        elif tssecmax: 
            conditions.append('(tssec <= %d)'%tssecmax)
        conditionStr = '&'.join(conditions) 
        return conditionStr

class dataRangeIterator:    
    def __init__(self,filehandles,tablenames,conditionStr):
        self._filehandles = filehandles
        self._tablenames = tablenames
        self._conditionStr = conditionStr

    def _get_range_in_file(self,filehandle):
        '''
        Selected range of all tables in file
        Output: {tablename:coordinates}
        '''
        result = {}
        if not self._conditionStr:
            return result
        for tablename in self._tablenames:
            n = filehandle.get_node('/'+tablename)
            coordinates = n.get_where_list(self._conditionStr,sort=False)
            if coordinates.size == 0: 
                result[tablename] = None
            else:
                result[tablename] = coordinates.tolist()
        return result
 
    def next(self):
        '''
        Output: (filehandle, {tablename: coordinates})
        '''
        for f in self._filehandles: #loop over files
            log.debug('Processing file '+f.filename)
            alltables = list(f.root._v_children.keys())
            if not _is_subset(self._tablenames,alltables):
                log.debug('file %s does not contain all the tables, skip '%f.filename)
                continue
            results = self._get_range_in_file(f)
            if not results:
                continue
            yield (f,results) 

    
def dataFilter(tablehandle,coordinates,field_dtypes,runlsRangeSelectSeries=None,runlsnbPointFilterSeries=None):
    '''
    Filter input table nodes on fields based on runls range selection and runlsnb point selection conditions
    Input:  
            tablehandle: table handle
            coordinates: row coordinates
            fields: output fields dtype
            runlsRangeSelectSeries: run,ls range selection matrix. Note: if None, select all
            runlsnbPointSelectSeries: [run,(ls,nb)]. Note: if None, select all
    Output: 
            numpy record array of type field_dtypes
    Note: this filter does not produce None. Empty result means reject all
    '''   
    myrows = tablehandle.read_coordinates(coordinates)
    thisrun = np.array(myrows['runnum'])[0]
    lsnums = np.array(myrows['lsnum'])
    masks = np.full(lsnums.shape, True, dtype=bool) #default to select all

    if runlsRangeSelectSeries is not None:
        selectedruns = runlsRangeSelectSeries.index
        if thisrun not in selectedruns: #run is not selected
            masks = np.full(lsnums.shape, False, dtype=bool) #mask out all ls
        else: 
            thisselectedls = runlsRangeSelectSeries[thisrun]
            masks = np.logical_or.reduce([np.logical_and(lsnums>=xmin, lsnums<=xmax) for [xmin,xmax] in thisselectedls]) #find masks   

    if runlsnbPointFilterSeries is not None:
        if thisrun not in runlsnbPointFilterSeries.index: #run is not selected
            masks = np.full(lsnums[masks].shape, False, dtype=bool) #mask out all the rest ls                
        else:
            lsnbselect = runlsnbPointFilterSeries[thisrun].values.tolist() #this run values : [(lsnum,nbnum)] 
            mylsnbs = [tuple(x) for x in np.column_stack((myrows['lsnum'][masks],myrows['nbnum'][masks]))]
            for i,lsnb in enumerate(mylsnbs):
                if lsnb not in lsnbselect:
                    masks[i] = False
    all_fields = []
    for fieldname in field_dtypes.names:
        value = myrows[fieldname][masks]
        all_fields.append(value)
    nrows = masks[np.nonzero(masks)].size
    log.debug('selected rows %d'%nrows)
    result = np.core.records.fromarrays(all_fields,names=field_dtypes.names,dtype=field_dtypes)    
    return result

def andFilter(irecordsSize,conditions):
    masks = np.full(irecordsSize, True, dtype=bool) #default to select all      
    for co in conditions:
        #print 'passed n masks:', masks[np.nonzero(masks)].size
        masks = np.logical_and(masks,co)
    return masks


class typebuilder:   
    def __init__(self,datatablename,datatype,withBX): 
        '''
        Build np.dtype for input tables and output data
        Input: 
            datatablename: table name in hdf5 file
            withBX: if output is with per bx data       
        '''
        self._basetypelist = [('runnum','uint32'),('lsnum','uint32'),('nbnum','uint32'),('timestampsec','uint32')]
        self._tcdstypelist = self._basetypelist + [('cmson','bool8'),('deadfrac','float32'),('ncollidingbx','uint32')]
        self._beam_typelist = self._basetypelist + [('status','U28'),('machinemode','U20'),('targetegev','uint32')] 
        self._withBX = withBX  
        self._datatablename = datatablename
        self._datatype = datatype

    def datatablename(self):
        return self._datatablename

    def datatype(self):
        return self._datatype 

    def tcds_typelist(self):
        '''
            tcds table input fields 
        '''
        return np.dtype(self._tcdstypelist)

    def beam_typelist(self):
        '''
             beam table input fields 
        '''
        return np.dtype(self._beam_typelist)

    def lumidata_typelist(self):
        '''
            lumi data table input fields            
        '''
        if self._datatablename == 'bestlumi':
            datafieldtypelist = self._basetypelist + [('fillnum','uint32'),('delivered','float32'),('recorded','float32'),('avgpu','float32'),('provider','U8')]
            if self._withBX:
                datafieldtypelist.append(('bxdelivered','float32',(3564,)))
        else:
            if self._datatype == 'raw':
                datafieldtypelist = self._basetypelist + [('fillnum','uint32'),('avgraw','float32')]
                if self._withBX:
                    datafieldtypelist.append(('bxraw','float32',(3564,)))
            else:
                datafieldtypelist = self._basetypelist + [('fillnum','uint32'),('avg','float32')]
                if self._withBX:
                    datafieldtypelist.append(('bx','float32',(3564,)))
        return np.dtype(datafieldtypelist)
    
def resultIter(filehandles,lumitype,datatype='best',runmin=None,runmax=None,fillmin=None,tssecmin=None,tssecmax=None,fillmax=None,beamstatusid=None,amodetagid=None,targetegev=None,runlsselect=None,withBX=False):
    '''
    get online bestlumi
    field choices: [runnum,lsnum,fillnum,timestampsec,cmson,beamstatus,delivered,recorded,bx,avgpu,datasource,normtag,normtagid,amodetagid,targetegev,numbxbeamactive,norb,nbperls]
    
    '''
    datatablename = params._lumitypeToh5tablename[lumitype.upper()]
    tbuilder = typebuilder(datatablename,datatype,withBX)
    preconditionStr = _build_preselectcondition(runmin=runmin,runmax=runmax,fillmin=fillmin,tssecmin=tssecmin,tssecmax=tssecmax,fillmax=fillmax,runlsselect=runlsselect)
    rangeIter = dataRangeIterator(filehandles,['tcds','beam',datatablename],preconditionStr)
         
    for co in rangeIter.next():
        filehandle = co[0]
        all_coordinates = co[1].values()
        if None in all_coordinates:
            log.debug('not all tables passed preselection, skip file ',filehandle.filename)
            continue

        #tcds selection

        tcds_coordinates = co[1]['tcds']
        tcds_table_handle = filehandle.get_node('/tcds')
        tcds_result = dataFilter(tcds_table_handle,tcds_coordinates,field_dtypes=tbuilder.tcds_typelist(),runlsRangeSelectSeries=runlsselect,runlsnbPointFilterSeries=None)

        if not tcds_result.size:
            log.debug('tcds failed data selection, continue ',filehandle.filename)
            continue
     
        selected_time = np.column_stack((tcds_result['runnum'],tcds_result['lsnum'],tcds_result['nbnum']))       
        runlsnbSeries = _make_runlsnb_Series( selected_time )
        #beam selection

        beam_coordinates = co[1]['beam']
        beam_table_handle = filehandle.get_node('/beam')
        beam_result = dataFilter(beam_table_handle,beam_coordinates,field_dtypes=tbuilder.beam_typelist(),runlsRangeSelectSeries=None,runlsnbPointFilterSeries=runlsnbSeries)
        beamconditions = []
        if beamstatusid is not None:
            req_beamstatus = params._idtobeamstatus[beamstatusid]
            beamconditions.append( beam_result['status']==req_beamstatus )
        if amodetagid is not None:
            req_amodetag = params._idtoamodetag[amodetagid]
            if params._amodetagtofull.has_key(req_amodetag):
                req_amodetag = params._amodetagtofull[req_amodetag]
            beamconditions.append( beam_result['machinemode']==req_amodetag )
        if targetegev is not None:
            beamconditions.append( beam_result['targetegev']==targetegev )
        beam_masks = andFilter(beam_result.size,beamconditions)        
        beam_result = beam_result[beam_masks]

        if not beam_result.size:
            log.debug('beam failed data selection, continue ',filehandle.filename)
            continue

        beam_time = np.column_stack((beam_result['runnum'],beam_result['lsnum'],beam_result['nbnum']))  
        if not np.array_equal(beam_time,selected_time):
            log.debug('beam time differs from tcds time, remake a narrower selection')
            runlsnbSeries = _make_runlsnb_Series( beam_time)

        #data selection

        data_coordinates = co[1][datatablename]
        data_table_handle = filehandle.get_node('/'+datatablename)

        data_result = dataFilter(data_table_handle,data_coordinates,field_dtypes=tbuilder.lumidata_typelist(),runlsRangeSelectSeries=None,runlsnbPointFilterSeries=runlsnbSeries)      

        for record in data_result:
            r = {}
            r['fillnum'] = record['fillnum']
            r['runnum'] = record['runnum']
            r['lsnum'] = lsnum = record['lsnum']
            nbnum = record['nbnum']
            r['timestampsec'] = record['timestampsec']
            tcds_masks = andFilter(tcds_result.size,[tcds_result['lsnum']==lsnum , tcds_result['nbnum']==nbnum])
            tcds_data = tcds_result[tcds_masks]
            r['deadtimefrac'] = tcds_data['deadfrac'][0]
            r['cmson'] = tcds_data['cmson'][0]
            r['numbxbeamactive'] = tcds_data['ncollidingbx'][0]
            beam_data = beam_result[ np.logical_and(beam_result['lsnum']==lsnum , beam_result['nbnum']==nbnum) ]            
            r['beamstatusid'] = params._beamstatustoid[beam_data['status'][0]]
            machinemode = beam_data['machinemode'][0]
            r['amodetag'] = params._fulltoamodetag[machinemode]
            r['targetegev'] = beam_data['targetegev']

            if datatype == 'best':
                r['delivered'] = record['delivered']
                r['recorded'] = record['recorded']
                r['avgpu'] = record['avgpu']
                r['datasource'] = record['provider']
                if withBX:
                    r['bxdeliveredblob'] = record['bxdelivered']
            elif datatype == 'raw':        
                r['avglumi'] = record['avgraw']             
                if withBX:
                    r['bxlumiblob'] = record['bxraw']
            else:
                r['avglumi'] = record['avg']
                if withBX:
                    r['bxlumiblob'] = record['bx']            
            yield r 
def _make_runlsnb_Series(runlsnbarray):
    '''
    Make pandas Series from [[run,ls,nb]] structure
    Input: 2-D structure [[run,ls,nb]]
    Output: pandas Series [(ls,nb), index=runs]
    '''
    runs = runlsnbarray[:,:1]
    lsnb = [tuple(x) for x in runlsnbarray[:,1:3]]
    return pd.Series(lsnb,index=runs.T[0])

if __name__=='__main__':
    tables = ['beam','tcds','hfetlumi']
    #runlsselect = pd.Series([[1340,1345],[1400,1500]], index=[327554,327554]) #pd.Series , index=runnum, value=[[lsmin,lsmax]]
    #runlsselect = pd.Series([[1,10],[14,15]], index=[327560,327560]) #pd.Series , index=runnum, value=[[lsmin,lsmax]]
    runlsselect = None

    #filenames = ['/home/zhen/data/7491/7491_327554_1812020507_1812020558.hd5','/home/zhen/data/7491/7491_327559_1812020558_1812020731.hd5','/home/zhen/data/7491/7491_327560_1812020731_1812021237.hd5']
    
    filenames =  ['/home/zhen/data/7491/7491_327560_1812020731_1812021237.hd5']
    lumitype = 'BEST'
    filehandles = open_validfiles(filenames,lumitype)
    print(filehandles)
    for result in resultIter(filehandles,lumitype,'best',fillmin=7491,fillmax=7491,runlsselect=runlsselect,beamstatusid=11,withBX=True):
        print(result['fillnum'],result['runnum'],result['lsnum'],result['timestampsec'],result['beamstatusid'],result['cmson'],result['delivered'],result['recorded'])
    lumitype = 'HFOC'
    for result in resultIter(filehandles,lumitype,'raw',fillmin=7491,fillmax=7491,runlsselect=runlsselect,beamstatusid=11,withBX=False):
        print(result['fillnum'],result['runnum'],result['lsnum'],result['timestampsec'],result['beamstatusid'],result['cmson'],result['avglumi'])

    [f.close() for f in filehandles] 
    




from __future__ import print_function
import sys
import os
sys.path.insert(0,os.path.dirname(sys.executable)+'/../lib/python2.7/site-packages/') #ignore other PYTHONPATH

import logging
import docopt
import schema
import brilws
import prettytable
import pandas as pd
import numpy as np
import ast
from brilws import api,params,display,lumiParameters,corrector
from brilws.cli import clicommonargs
import brilws.formatter as fm
import re,time, csv
from datetime import datetime
from sqlalchemy import *
import math
from dateutil import tz
import pytz

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

log = logging.getLogger('brilws')
logformatter = logging.Formatter('%(levelname)s %(name)s %(message)s')
log.setLevel(logging.ERROR)
ch = logging.StreamHandler()
ch.setFormatter(logformatter)
log.addHandler(ch)

lumip = lumiParameters.ParametersObject()
lslengthsec= lumip.lslengthsec
utctmzone = tz.gettz('UTC')
cerntmzone = tz.gettz('CEST')

class Unbuffered(object):
    '''
    replace sys.stdout with some other stream-like wrapper which does a flush after every call.
    '''
    def __init__(self, stream):
        self.stream = stream
    def write(self, data):
        self.stream.write(data)
        self.stream.flush()
    def __getattr__(self, attr):
        return getattr(self.stream,attr)
sys.stdout = Unbuffered(sys.stdout)

np.seterr(divide='ignore', invalid='ignore')

def xing_indexfilter(arr,constfactor=1.,xingMin=0.,xingTr=0.,xingId=[]):
    vidx = None
    if(xingMin is None):
        xingMin = float('-inf')
    if xingTr:        
        vidx = np.argwhere( np.logical_and( (arr*constfactor)>xingMin, arr>xingTr*np.max(arr) ) ).ravel()
    else:        
        vidx = np.argwhere( (arr*constfactor)>xingMin ).ravel()
    if xingId :
        posidx = np.array(xingId)-1 #array position index is bunch index - 1
        return np.intersect1d(np.array(posidx),vidx)    
    return vidx

def totalprescaleNEW(hltprescval,l1seedlogic,l1prescvals):
    '''
    1) if ONE single seed, take it
    2) if OR:  (A.1 OR A.2 OR A.3 ...) ignore the triggers with prescale 0, and take the minimum of the remaining prescale values:
        if max(A.i) == 0
                A = 0
        else
                A = min(A.i | A.i > 0)
    3) if AND: A AND B AND C ... take the product of the prescale values A * B * C ...
    4) (optional)
       if (A.1 OR A.2 OR A.3 ...) AND (B.1 OR B.2 OR ...) AND ...
       - for each group, take the lowest non-zero prescale, or 0 if they are all zero
         if max(A.i) = 0
                A =
         else
                A = min(A.i | A.i > 0)
       - take the product of the prescales thus obtained:
         P = A * B * ...
       users still take their own risk and should use their own judgement.
    5)  if NOT, or a combination of AND/OR and NOT, reject
    '''
    totpresc = 0
    if not hltprescval or not l1prescvals or not l1seedlogic:
        return totpresc
    if np.all( np.array(l1prescvals)==1 ):
        totpresc = hltprescval        
    elif l1seedlogic=='ONE':
        totpresc = hltprescval*l1prescvals[0]
    elif l1seedlogic=='OR':
        if np.max( np.array(l1prescvals) ) == 0:
            return totpresc
        else:
            l = np.array(l1prescvals)
            l1p = np.min( l[l>0] )
        totpresc = hltprescval*l1p
    elif l1seedlogic=='AND':
        if np.max( np.array(l1prescvals) ) == 0:
            return totpresc
        l1p = np.prod( np.array(l1prescvals) )
        totpresc = hltprescval*l1p    
    return totpresc

    
def totalprescale(hltprescval,l1seedlogic,l1prescvals):
    totpresc = 0
    if not hltprescval or not l1prescvals or not l1seedlogic:
        return totpresc
    if np.all( np.array(l1prescvals)==1 ):
        totpresc = hltprescval
    elif l1seedlogic=='ONE':
        totpresc = hltprescval*l1prescvals[0]
    elif l1seedlogic=='OR':
        
        totpresc = hltprescval*np.min(l1prescvals)
    elif l1seedlogic=='AND':
        totpresc = hltprescval*np.max(l1prescvals)    
    return totpresc

def lumi_per_normtag(shards,lumiquerytype,dbengine,dbschema,runtot,formatter,datasource=None,normtag=None,withBX=False,byls=None,fh=None,csvwriter=None,ptable=None,scalefactor=1,totz=utctmzone,fillmin=None,fillmax=None,runmin=None,runmax=None,amodetagid=None,egev=None,beamstatusid=None,tssecmin=None,tssecmax=None,runlsSeries=None,hltl1map={},ignorel1mask=False,xingMin=0.,xingTr=0.,xingId=[],checkjson=False,datatagnameid=None,withfileinput = False,dataset=None):
   
    if withfileinput:
        from brilws import fileapi

    validitychecker = None
    lastvalidity = None
    if normtag and normtag is not 'withoutcorrection':
        normdata = api.iov_gettagdata(dbengine, normtag,schemaname=dbschema)
        if not normdata: raise ValueError('normtag %s does not exist'%normtag)
        validitychecker = ValidityChecker(normdata)

    for shard in shards:              
        lumiiter = None
        if lumiquerytype == 'bestresultonline':
            if withfileinput:
               lumiiter = fileapi.resultIter(shards,datasource,fillmin=fillmin,fillmax=fillmax,runmin=runmin,runmax=runmax,amodetagid=amodetagid,targetegev=egev,beamstatusid=beamstatusid,tssecmin=tssecmin,tssecmax=tssecmax,runlsselect=runlsSeries,withBX=withBX)
            else:
                tablename = 'online_result_'+str(shard)
                shardexists = api.table_exists(dbengine,tablename,schemaname=dbschema)
                if not shardexists: continue
                rfields = ['fillnum','runnum','lsnum','timestampsec','cmson','beamstatusid','targetegev','delivered','recorded','avgpu','datasource','numbxbeamactive']
                if withBX: rfields = rfields+['bxdeliveredblob']                    
                lumiiter = api.online_resultIter(dbengine,tablename,schemaname=dbschema,fields=rfields,fillmin=fillmin,fillmax=fillmax,runmin=runmin,runmax=runmax,amodetagid=amodetagid,targetegev=egev,beamstatusid=beamstatusid,tssecmin=tssecmin,tssecmax=tssecmax,runlsselect=runlsSeries,sorted=True)
            
        elif lumiquerytype in ['detresultonline','detraw']:   
            datatype = 'raw'
            if lumiquerytype=='detresultonline':
                datatype = 'result'
            if withfileinput:
                lumiiter = fileapi.resultIter(shards,datasource,datatype=datatype,fillmin=fillmin,fillmax=fillmax,runmin=runmin,runmax=runmax,amodetagid=amodetagid,targetegev=egev,beamstatusid=beamstatusid,tssecmin=tssecmin,tssecmax=tssecmax,runlsselect=runlsSeries,withBX=withBX)   
            else:                
                tablename = datasource.lower()+'_'+datatype+'_'+str(shard)
                shardexists = api.table_exists(dbengine,tablename,schemaname=dbschema)
                if not shardexists: continue
                rfields = ['avglumi']
                idfields = ['fillnum','runnum','lsnum','timestampsec','beamstatusid','cmson','deadtimefrac','targetegev','numbxbeamactive']
                if withBX: 
                    rfields = rfields+['bxlumiblob']
                lumiiter = api.dataIter(dbengine,datasource,datatype,shard,datafields=rfields,idfields=idfields,schemaname=dbschema,fillmin=fillmin,fillmax=fillmax,runmin=runmin,runmax=runmax,amodetagid=amodetagid,targetegev=egev,beamstatusid=beamstatusid,tssecmin=tssecmin,tssecmax=tssecmax,runlsselect=runlsSeries,sorted=True,datatagnameid=datatagnameid)                    
                  
        if lumiiter is None:
            continue
        
        g_run_old = 0
        g_ls_trglastscaled_old = 0
        prescale_map = {} # for global scope
        g_hltconfigid = 0
        for row in lumiiter:
            fillnum = row['fillnum']
            runnum = row['runnum']
            lsnum = row['lsnum']            
            cmslsnum = lsnum
            timestampsec = row['timestampsec']
            cmson = row['cmson']
            
            if not cmson:
                cmslsnum = 0
            if hltl1map:
                if cmslsnum==0: continue  #cms is not running, skip.                
                if g_run_old!=runnum:     #on new run boundary, get hltconfigid
                    hltrunconfig = api.get_hltrunconfig(dbengine,runnum=runnum,schemaname=dbschema) #[['runnum','hltconfigid','hltkey']]
                    if not hltrunconfig:
                        continue
                    g_hltconfigid = [h[1] for h in hltrunconfig][0]
                    if not g_hltconfigid:
                        continue      
                    presc = api.get_prescidx_change(dbengine,runnum,schemaname=dbschema)#{runnum:[[lslastscaler,prescidx]] }
                    if runnum in presc:
                        presc = presc[runnum]
                this_prescidx = None
                if not presc:                    
                    ls_trglastscaled = 1
                else:
                    b = [ i[0] for i in presc if i[0]<=cmslsnum ] 
                    if not b:
                        ls_trglastscaled = 1
                        this_prescidx = None                        
                    else:
                        ls_trglastscaled = np.max( b ).item()
                        this_prescidx = [t[1] for t in presc if t[0]==ls_trglastscaled][0]
                if (g_ls_trglastscaled_old != ls_trglastscaled) or (g_run_old != runnum): #on prescale change lumi section or new run
                    prescale_map = {} #clear prescaleindex change map on new prescale index value
                    if g_hltconfigid not in hltl1map:
                        continue
                    this_hltl1map = hltl1map[g_hltconfigid]#[[hltpathid,hltpathname,l1seedtype,l1seedbits]]
                    for grouped in this_hltl1map:
                        this_hltpathid = grouped[0]
                        hltpathname = grouped[1]
                        l1seedlogic = grouped[2]
                        l1candidates = grouped[3]
                        if this_prescidx is None:
                            hltprescval = 1
                            totpresc = 1
                        else:
                            if dataset is None:
                              hltprescval = api.get_hltprescale(dbengine,runnum,ls_trglastscaled,g_hltconfigid,this_prescidx,this_hltpathid,schemaname=dbschema)
                            else:
                              hltprescval = api.get_hltprescale_in_dataset(dbengine,runnum,ls_trglastscaled,g_hltconfigid,this_prescidx,this_hltpathid,dataset,schemaname=dbschema)
                            rl = api.get_l1prescale(dbengine,runnum,ls_trglastscaled,l1candidates=l1candidates,prescidxs=this_prescidx,ignorel1mask=ignorel1mask ,schemaname=dbschema)
                            if not rl: continue
                            l1keys = [ k for k in rl.keys() if k[0]==this_prescidx ]
                            l1vals = [ rl[k] for k in l1keys ]                       
                            l1prescvals = [ v[0] for v in l1vals ]                            
                            totpresc = totalprescaleNEW(hltprescval,l1seedlogic,l1prescvals)
                            if not totpresc:
                                continue
                        prescale_map[hltpathname] = totpresc   
                    g_ls_trglastscaled_old = ls_trglastscaled
                g_run_old = runnum
            beamstatusid = row['beamstatusid']
            beamstatus = params._idtobeamstatus[beamstatusid]
            if beamstatus not in ['FLAT TOP','STABLE BEAMS','SQUEEZE','ADJUST']: continue
            tegev = row['targetegev']
            dtime = str(timestampsec)
            if totz is not None:
                d = datetime.fromtimestamp(int(timestampsec),tz=pytz.utc)
                dtime = d.astimezone(totz).strftime(params._datetimefm)
            delivered = recorded = avgpu = livefrac = 0.
            if checkjson:
                g_returnedls.append((runnum,lsnum))
            if lumiquerytype == 'bestresultonline': ##bestlumi                                
                if 'delivered' in row and row['delivered']:                    
                    delivered = np.true_divide(row['delivered']*lslengthsec,scalefactor)             
                if delivered>0 and 'recorded' in row and row['recorded']:
                    recorded = np.true_divide(row['recorded']*lslengthsec,scalefactor)
                if delivered>0 and 'avgpu' in row and row['avgpu']:
                    avgpu = row['avgpu']
                ds = 'UNKNOWN' 
                if 'datasource' in row and row['datasource']:
                    ds = row['datasource']
                livefrac = np.true_divide(recorded,delivered)
                if withBX:    #--xing
                    bxlumi = None
                    bxlumistr = '[]'
                    
                    if 'bxdeliveredblob' in row and row['bxdeliveredblob'] is not None:
                        if withfileinput:
                            bxdeliveredarray = row['bxdeliveredblob']
                        else:
                            bxdeliveredarray = np.array(api.unpackBlobtoArray(row['bxdeliveredblob'],'f'))                        
                        totfactor = np.true_divide(lslengthsec,scalefactor)
                        bxidx = xing_indexfilter(bxdeliveredarray,constfactor=totfactor,xingMin=xingMin,xingTr=xingTr,xingId=xingId)
                        if bxidx is not None and bxidx.size>0:
                            bxdelivered = bxdeliveredarray[bxidx]*totfactor
                            bxlumi = np.transpose( np.array([bxidx+1,bxdelivered,bxdelivered*livefrac]) )
                        del bxdeliveredarray
                        del bxidx
                        if hltl1map: #--hltpath bx display
                            for pth in prescale_map.keys():                                
                                thispresc = np.float64( prescale_map[pth] )
                                if bxlumi is not None:
                                    a = map(formatter.bxlumi,bxlumi/thispresc)  
                                    bxlumistr = '['+' '.join(a)+']'                                
                                display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,pth,formatter.lumi(np.true_divide(delivered,thispresc)),formatter.lumi(np.true_divide(recorded,thispresc)),avgpu,ds,'%s'%bxlumistr] , fh=fh, csvwriter=csvwriter, ptable=ptable)                                
                        else:        #normal bx display                   
                            if bxlumi is not None:
                                a = map(formatter.bxlumi,bxlumi)
                                bxlumistr = '['+' '.join(a)+']'                            
                            display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,beamstatus,'%d'%tegev,formatter.lumi(delivered),formatter.lumi(recorded),'%.1f'%(avgpu),ds,'%s'%bxlumistr] , fh=fh, csvwriter=csvwriter, ptable=ptable)                            
                    del bxlumi
                    
                elif byls:      #--byls
                    if hltl1map:#--hltpath display
                        for pth in prescale_map.keys():
                            thispresc = np.float64( prescale_map[pth] )
                            display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum), dtime, pth, formatter.lumi(np.true_divide(delivered,thispresc)),formatter.lumi(np.true_divide(recorded,thispresc)),'%.1f'%(avgpu),ds] , fh=fh, csvwriter=csvwriter, ptable=ptable)                            
                    else:       #normal display
                        display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,beamstatus,'%d'%tegev,formatter.lumi(delivered),formatter.lumi(recorded),'%.1f'%(avgpu),ds] , fh=fh, csvwriter=csvwriter, ptable=ptable)                        
            else:               ###lumisource
                if 'deadtimefrac' in row :
                    if row['deadtimefrac'] is not None:
                        livefrac = 1.-row['deadtimefrac']
                    else:
                        if cmslsnum!=0:
                            g_nulldeadtime.setdefault(runnum,[]).append(lsnum)
                        livefrac = 0.
                uncorrectedavglumi = row['avglumi'] #uncorrected, hz/ub
                avglumi = uncorrectedavglumi
                ncollidingbx = row['numbxbeamactive']
                uncorrectedbxlumi = None
                if 'bxlumiblob' in row and row['bxlumiblob'] is not None:                    
                    if withfileinput:
                        uncorrectedbxlumi = row['bxlumiblob']
                    else:
                        uncorrectedbxlumi = np.array(api.unpackBlobtoArray(row['bxlumiblob'],'f'))                
                bxlumi = uncorrectedbxlumi
                if validitychecker is not None:
                    if not lastvalidity or not validitychecker.isvalid(runnum,lastvalidity):
                        lastvalidity = validitychecker.getvalidity(runnum)
                    [normfunc,normparam] = validitychecker.getvaliddata(lastvalidity[0])
                    if withBX:
                        (avglumi,bxlumi) = corrector.applyCorrectionStr( normfunc,normparam, corrector.FunctionRoot(uncorrectedbxlumi,avglumi,ncollidingbx) )
                    else:
                        avglumi = corrector.applyCorrectionStr( normfunc,normparam, corrector.FunctionRoot(avglumi,avglumi,ncollidingbx) )[0]

                totfactor = np.true_divide(lslengthsec,scalefactor)
                delivered = avglumi*totfactor
                recorded = delivered*livefrac
                avgpu = lumip.avgpu( avglumi,ncollidingbx,g_minbias)

                if withBX:      #--xing
                    bxlumistr = '[]'
                    if bxlumi is not None:
                        bxidx = xing_indexfilter(bxlumi,constfactor=totfactor,xingMin=xingMin,xingTr=xingTr,xingId=xingId)          
                        if bxidx is not None and bxidx.size>0:           
                            bxdelivered =  bxlumi[bxidx]*totfactor
                            bxlumi = np.transpose( np.array([bxidx+1,bxdelivered,bxdelivered*livefrac]) )
                        else:
                            bxlumi = None                       
                        del uncorrectedbxlumi
                        del bxidx
                    if hltl1map:#--hltpath xing display
                        for pth in prescale_map.keys():
                            thispresc = np.float64( prescale_map[pth] ) 
                            if bxlumi is not None:
                                a = map(formatter.bxlumi,bxlumi/thispresc)  
                                bxlumistr = '['+' '.join(a)+']'               
                            display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,pth,formatter.lumi(np.true_divide(delivered,thispresc)),formatter.lumi(np.true_divide(recorded,thispresc)),avgpu,datasource.upper(),'%s'%bxlumistr] , fh=fh, csvwriter=csvwriter, ptable=ptable)                            
                    else:       #normal xing display
                        if bxlumi is not None:
                            a = map(formatter.bxlumi,bxlumi)                            
                            bxlumistr = '['+' '.join(a)+']'
                        display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,beamstatus,'%d'%tegev,formatter.lumi(delivered),formatter.lumi(recorded),'%.1f'%(avgpu),datasource.upper(),'%s'%bxlumistr] , fh=fh, csvwriter=csvwriter, ptable=ptable)                        
                    del bxlumi
                elif byls:       #--byls
                    if hltl1map: #--hltpath display
                        for pth in prescale_map.keys():
                            thispresc = np.float64(prescale_map[pth])
                            display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,pth,formatter.lumi(np.true_divide(delivered,thispresc)),formatter.lumi(np.true_divide(recorded,thispresc)),'%.1f'%(avgpu),datasource.upper()] , fh=fh, csvwriter=csvwriter, ptable=ptable)                            
                    else:        #normal display
                        display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,beamstatus,'%d'%tegev,formatter.lumi(delivered),formatter.lumi(recorded),'%.1f'%(avgpu),datasource.upper()] , fh=fh, csvwriter=csvwriter, ptable=ptable)    
            if not hltl1map:     #normal statistic collect
                if ('',runnum) not in runtot:
                    runtot[('',runnum)] = {'fill':fillnum,'dtime':dtime,'nls':0,'ncms':0,'delivered':0,'recorded':0} 
                runtot[ ('',runnum)]['nls']+=1                
                if cmson: runtot[ ('',runnum) ]['ncms']+=1
                runtot[ ('',runnum) ]['delivered']+=delivered
                runtot[ ('',runnum) ]['recorded']+=recorded                                
            else:                #hltpath statistic collect
                if not cmson:
                    continue
                for pth in prescale_map.keys():
                    thispresc = np.float64(prescale_map[pth])
                    if (pth,runnum) not in runtot:
                        runtot[(pth,runnum)] = {'fill':fillnum,'dtime':dtime,'nls':0,'ncms':0,'delivered':0,'recorded':0}                        
                    runtot[ (pth,runnum)]['nls']+=1                
                    if cmson: runtot[ (pth,runnum) ]['ncms'] += 1
                    runtot[ (pth,runnum) ]['delivered'] += np.true_divide(delivered,thispresc )
                    runtot[ (pth,runnum) ]['recorded'] += np.true_divide(recorded,thispresc )      
                
class ValidityChecker(object):
    def __init__(self, normdata):
        self.normdata = normdata
        self.allsince = np.array([x[0] for x in normdata])
        
    def getvalidity(self,runnum):
        if self.allsince.size == 0 : 
            return None
        since = 1
        if self.allsince[self.allsince<=runnum].size>0:
            since = self.allsince[self.allsince<=runnum].max()
        till = 999999
        if self.allsince[self.allsince>runnum].size>0:
            till = self.allsince[self.allsince>runnum].min()
        return (since,till)
    
    def isvalid(self,runnum,validity):
        if runnum>=validity[0] and runnum<validity[1]:
            return True
        return False

    def getvaliddata(self,since):
        s = np.where(self.allsince==since)
        if s[0].size>0:
            sinceindex = s[0][0]
            func = self.normdata[sinceindex][1]
            params = self.normdata[sinceindex][2]
            return [func,params]
        return None

def findtagname(dbengine,datatagname,dbschema):
    '''
    output: (datatagname,datatagnameid)
    '''
    datatagnameid=0
    if datatagname:
        datatagnameid = api.datatagnameid(dbengine,datatagname=datatagname,schemaname=dbschema)
    else:
        r = api.max_datatagname(dbengine,schemaname=dbschema)
        if not r:
            raise RuntimeError('no tag found')
        datatagname = r[0]
        datatagnameid = r[1]
    if not datatagnameid:
        raise RuntimeError('No tag %s found'%datatagname)   
    return (datatagname,datatagnameid)

          
def brilcalc_main(progname=sys.argv[0]):

    docstr='''

    usage:
      brilcalc (-h|--help|--version) 
      brilcalc [--debug|--warn] <command> [<args>...]

    commands:
      lumi Luminosity
      beam Beam 
      trg Trigger configurations      
    See 'brilcalc <command> --help' for more information on a specific command.

    '''
    args = {}
    argv = sys.argv[1:]
    args = docopt.docopt(docstr,argv,help=True,version=brilws.__version__,options_first=True)

    if args['--debug']:
        log.setLevel(logging.DEBUG)
    elif args['--warn']:
        log.setLevel(logging.WARNING)
        
    log.debug('global arguments: %s',args)
    cmmdargv = [args['<command>']] + args['<args>']

    log.debug('command arguments: %s',cmmdargv)
    parseresult = {}
    global g_returnedls #[(run,ls),...]
    global g_nulldeadtime #{run:[]}
    global g_minbias #minbias xsec    
    
    try:
      if args['<command>'] == 'lumi':
          from . import brilcalc_lumi          
          parseresult = docopt.docopt(brilcalc_lumi.__doc__,argv=cmmdargv)
          parseresult = brilcalc_lumi.validate(parseresult)
          if parseresult['--filedata'] is not None:
              if '-u' in parseresult and parseresult['-u']!= 'hz/ub':
                  print('[WARN] output unit is hz/ub with option --filedata')
              parseresult['-u'] = 'hz/ub'
          ##parse selection params
          pargs = clicommonargs.parser(parseresult)
          dbschema = ''
          if not pargs.dbconnect.find('oracle')!=-1: dbschema = 'cms_lumi_prod'
          log.debug('connecturl: %s'%pargs.connecturl)
          if sys.version_info > (3,):
              dbengine = create_engine(pargs.connecturl,max_identifier_length=128)
          else:
              dbengine = create_engine(pargs.connecturl)
          totz=utctmzone
          if pargs.cerntime:
              totz=cerntmzone
          elif pargs.tssec:
              totz=None          
        
          fillmin = pargs.fillmin
          fillmax = pargs.fillmax
          runmin = pargs.runmin
          runmax = pargs.runmax
          amodetagid = pargs.amodetagid
          egev = pargs.egev
          beamstatusid = pargs.beamstatusid
          tssecmin = pargs.tssecmin
          tssecmax = pargs.tssecmax
          selectionkwds = {}
          normtag = None          
          fh = None
          ptable = None
          ftable = None
          csvwriter = None
          myformat = pargs.oformat
          myprecision = pargs.precision
          myformatter = fm.Formatter(myformat,myprecision)
          vfunc_lumiunit = np.vectorize(myformatter.lumiunit)
          checkjson = True
          if parseresult['--without-checkjson'] or not parseresult['-i']:
              checkjson = False
          if parseresult['--normtag'] or parseresult['--type']:
              g_minbias = pargs.minbias or lumip.minbias
          else:
              if not pargs.minbias is None:
                  print('Error: --minBiasXsec cannot be used without --normtag 0r --type, quit')
                  sys.exit(0)
                  
          g_returnedls = []
          g_nulldeadtime = {}
          g_headers = {}
          g_headers['runheader'] = ['run:fill','time','nls','ncms','delivered(/ub)','recorded(/ub)']
          g_headers['footer'] = ['nfill','nrun','nls','ncms','totdelivered(/ub)','totrecorded(/ub)']
          g_headers['bylsheader'] = ['run:fill','ls','time','beamstatus','E(GeV)','delivered(/ub)','recorded(/ub)','avgpu','source']
          g_headers['runheader_hltpath'] = ['run:fill','time','ncms','hltpath','delivered(/ub)','recorded(/ub)']
          g_headers['footer_hltpath'] = ['hltpath','nfill','nrun','ncms','totdelivered(/ub)','totrecorded(/ub)']
          g_headers['bylsheader_hltpath'] = ['run:fill','ls','time','hltpath','delivered(/ub)','recorded(/ub)','avgpu','source']          

          
          scalefactor = pargs.scalefactor          
          lumiunitstr = parseresult['-u']         
          if lumiunitstr not in myformatter.lumiunit_to_scalefactor.keys(): 
              raise ValueError('%s not recognised as lumi unit'%lumiunit)
          lumiunitconversion = myformatter.lumiunit_to_scalefactor[lumiunitstr]
          scalefactor = scalefactor*lumiunitconversion
         
          if pargs.withBX:
              if pargs.hltpath is None:
                  header = g_headers['bylsheader']+['[bxidx bxdelivered(/ub) bxrecorded(/ub)]']
                  footer = g_headers['footer']
              else:
                  header = g_headers['bylsheader_hltpath']+['[bxidx bxdelivered(/ub) bxrecorded(/ub)]']
                  footer = g_headers['footer_hltpath']
          elif pargs.byls:
              if pargs.hltpath is None:
                  header = g_headers['bylsheader']
                  footer = g_headers['footer']
              else:
                  header = g_headers['bylsheader_hltpath']
                  footer = g_headers['footer_hltpath']
          else:
              if pargs.hltpath is None:
                  header = g_headers['runheader']
                  footer = g_headers['footer']
              else:
                  header = g_headers['runheader_hltpath']
                  footer = g_headers['footer_hltpath']
                  
          header = vfunc_lumiunit(header,lumiunitstr).tolist()
          footer = vfunc_lumiunit(footer,lumiunitstr).tolist()
        
          
          (datatagname,datatagnameid) = findtagname(dbengine,pargs.datatagname,dbschema)
          if not pargs.totable:
              fh = pargs.ofilehandle
              print('#Data tag : %s , Norm tag: %s'%(datatagname,normtag), file=fh)
              print('#'+','.join(header),file = fh)
              csvwriter = csv.writer(fh)
          else:
              ptable = display.create_table(header,header=True,align='l')
              ftable = display.create_table(footer,header=True,align='l')          
              
          datasources = [] #[lumiquerytype,normtagname,datasource,runlsstr]          

          lumiquerytype = 'detraw'
          normtag = normtagname = 'withoutcorrection'
          parsediffruns = None
          parsediffls = None
          if not pargs.withoutcorrection:
              normtag = pargs.iovtagSelect
              if not normtag:                  
                  if pargs.lumitype:
                      lumiquerytype = 'detresultonline'
                      datasources.append( ['detresultonline',normtag,pargs.lumitype.lower(),pargs.runlsSeries] )
                      normtagname = 'onlineresult'
                  else:
                      lumiquerytype = 'bestresultonline'
                      normtagname = 'onlineresult'
                      datasources.append( ['bestresultonline',normtag,'best',pargs.runlsSeries] )
              else:
                  if isinstance(normtag,list): #normtag is list
                      if pargs.runlsSeries is None:
                          checkjson = False
                          #mergedselect = normtag
                          mergedselect = api.mergeiovrunlsWithDB(dbengine,normtag,runmin=runmin,runmax=runmax,fillmin=fillmin,fillmax=fillmax,tssecmin=tssecmin,tssecmax=tssecmax,schemaname=dbschema)                          
                      else:
                          (parsediffruns, parsediffls) = api.checksuperset([x[1] for x in normtag],pargs.runlsSeries)                         
                          mergedselect = api.mergeiovrunls(normtag,pargs.runlsSeries)
                      normtagname = 'composite'
                      for item in mergedselect:
                          iovtag = item[0]
                          runlsdict = item[1]
                          iovtag_meta = api.iov_gettag(dbengine,iovtag,schemaname=dbschema)
                          if not iovtag_meta: raise ValueError('requested iovtags do not exist')  
                          datasource = iovtag_meta[2].lower()
                          datasources.append( [lumiquerytype,iovtag,datasource,runlsdict] )
                  else:  #normtag is string (iov tag)
                      datasource = pargs.lumitype
                      normtagname = normtag
                      if datasource is None: #det lumi with correction
                          iovtag_meta = api.iov_gettag(dbengine,normtag,schemaname=dbschema)
                          if not iovtag_meta: raise ValueError('%s does not exist'%normtag)                   
                          datasource = iovtag_meta[2]
                      datasources.append( [lumiquerytype,normtag,datasource.lower(),pargs.runlsSeries ])
                     
          else:
              if not pargs.lumitype:
                  raise ValueError('--type is required with --without-correction')
              datasources.append( [lumiquerytype,None,pargs.lumitype.lower(),pargs.runlsSeries ])
          
          log.debug('lumiquerytype %s'%lumiquerytype)          
          log.debug('scalefactor: %.2f'%pargs.scalefactor)                    
          
          runtot = {} #{(hltpath,run): { 'fill':fillnum,'time':dtime,'nls':1,'ncms':int(cmson),'delivered':delivered,'recorded':recorded} }
                               
          hltl1map = None #{hltconfigid:[[hltpathid,hltpathname,seedtype,[seedvalues]]]}
          if pargs.hltpath is not None:           
              hltl1map = api.get_hlttrgl1seedmap(dbengine,hltpath=pargs.hltpath,schemaname=dbschema)
              if not hltl1map:
                  print('no hltpath to l1bit mapping found')
                  sys.exit(0)
              if pargs.dataset is not None:
                  datasethltpairs = set()
                  for k in hltl1map: 
                    hpaths = [h[0] for h in hltl1map[k]]
                    for hp in hpaths:                        
                      datasethltpairs.add( (k,hp) )
                  for (hltconfigid,hltpathid) in datasethltpairs:
                    r = api.is_hltpathid_in_dataset(dbengine,hltpathid,pargs.dataset,hltconfigid,schemaname=dbschema)
                  ##check hltpathid and dataset are in the same menu
                    if not r:
                      print( 'dataset {} and hltpath{} are not in the same menu'.format(pargs.dataset,pargs.hltpath) )
                      sys.exit(0)

          rselectrange = [] 
          if datasources :
              for [qt,nt,ds,rs] in datasources:
                  if rs is None: 
                      continue
                  for r,l in rs.iteritems():
                      if not r in rselectrange:
                          rselectrange.append(r)   
          withfileinput = False
          if pargs.filedata is not None:
              withfileinput = True
              from brilws import fileapi
              if os.path.isfile(pargs.filedata):
                  filenames = [ os.path.abspath(pargs.filedata) ]
              else:
                  filenames = [ os.path.join(os.path.abspath(pargs.filedata),f) for f in os.listdir(pargs.filedata) ]
              shards = fileapi.open_validfiles(filenames,pargs.lumitype or 'best')              
              if not shards:
                  print('Failed to find data file for the requested time range.')
                  [f.close() for f in shards] 
                  sys.exit(1)
          else:          
              shards = api.locate_shards(dbengine,runmin=runmin,runmax=runmax,fillmin=fillmin,fillmax=fillmax,tssecmin=tssecmin,tssecmax=tssecmax,orrunlist=rselectrange,schemaname=dbschema)
              if not shards:
                  print('Failed to find data table for the requested time range.')
                  sys.exit(1)
              log.debug('shards: '+str(shards))
          
          for [qtype,ntag,dsource,rselect] in datasources:
              lumi_per_normtag(shards,qtype,dbengine,dbschema,runtot,myformatter,datasource=dsource,normtag=ntag,withBX=pargs.withBX,byls=pargs.byls,fh=fh,csvwriter=csvwriter,ptable=ptable,scalefactor=scalefactor,totz=totz,fillmin=fillmin,fillmax=fillmax,runmin=runmin,runmax=runmax,amodetagid=amodetagid,egev=egev,beamstatusid=beamstatusid,tssecmin=tssecmin,tssecmax=tssecmax,runlsSeries=rselect,hltl1map=hltl1map,ignorel1mask=parseresult['--ignore-mask'],xingMin=pargs.xingMin,xingTr=pargs.xingTr,xingId=pargs.xingId,checkjson=checkjson,datatagnameid=datatagnameid,withfileinput=withfileinput,dataset=pargs.dataset)  
          if pargs.filedata:
              [f.close() for f in shards] 
          if pargs.hltpath is None:
              nruns = len(runtot.keys())
              nfills = len( set([v['fill'] for v in runtot.values()] ) )
              nls = np.sum( [v['nls'] for v in runtot.values()]  )
              ncmsls = np.sum( [v['ncms'] for v in runtot.values()] )
              totdelivered = np.sum( [v['delivered'] for v in runtot.values() ] )
              totrecorded =  np.sum( [v['recorded'] for v in runtot.values() ] )
              if not pargs.byls and not pargs.withBX: #run table
                  for hn,rn in sorted(runtot.keys()):
                      v = runtot[ (hn,rn) ]                      
                      display.add_row( ['%d:%d'%(rn,v['fill']),v['dtime'],v['nls'],v['ncms'],myformatter.lumi(v['delivered']),myformatter.lumi(v['recorded']) ] , fh=fh, csvwriter=csvwriter, ptable=ptable)
              if pargs.filedata is None:
                  display.add_row( [ '%d'%nfills, '%d'%nruns, '%d'%nls,'%d'%ncmsls,myformatter.lumi(totdelivered),myformatter.lumi(totrecorded)], fh=fh, csvwriter=None, ptable=ftable)
              
          else:
              hltpathSummary = []
              runtot_df = pd.DataFrame.from_dict(runtot,orient='index')
              grouped = runtot_df.groupby(level=0) #group by hltpath
              totalpathdelivered = 0
              totalpathrecorded = 0
              for hn,g in grouped:
                  nruns = g.index.size
                  nfills = int(g['fill'].nunique())
                  ncmsls =  int(g['ncms'].sum())
                  totdelivered = g['delivered'].sum()
                  totrecorded = g['recorded'].sum()
                  if pargs.filedata is None:
                      display.add_row( [ hn, '%d'%nfills, '%d'%nruns, '%d'%ncmsls,myformatter.lumi(totdelivered),myformatter.lumi(totrecorded)], fh=fh, csvwriter=None, ptable=ftable)                                 
                  if not pargs.byls and not pargs.withBX: #hltpath, run table 
                      for i,v in g.iterrows():
                          pname = i[0]
                          trun = i[1]
                          tfill = v['fill']
                          ttime = v['dtime']
                          tncmsls = v['ncms']
                          tdelivered = v['delivered']
                          trecorded = v['recorded']
                          display.add_row( [ '%d:%d'%(trun,tfill), ttime, tncmsls, pname, myformatter.lumi(tdelivered), myformatter.lumi(trecorded)], fh=fh, csvwriter=csvwriter, ptable=ptable )
                  hltpathSummary.append([hn,nfills,nruns,ncmsls,totdelivered,totrecorded])
                  totalpathdelivered =  totalpathdelivered+totdelivered
                  totalpathrecorded = totalpathrecorded+totrecorded
                   
              del runtot_df
          
          if pargs.totable:  #to screen              
              print('#Data tag : %s , Norm tag: %s'%(datatagname,normtagname))
              display.show_table(ptable,pargs.outputstyle)
              if pargs.filedata is None:
                  print ("#Summary: ")
                  display.show_table(ftable,pargs.outputstyle)
              del ptable
              del ftable
              if pargs.hltpath:
                  print ('#Sum delivered : %s'%myformatter.lumi(totalpathdelivered))
                  print ('#Sum recorded : %s'%myformatter.lumi(totalpathrecorded))
          else:              #to file
              print('#Summary:',file = fh)
              print('#'+','.join(footer),file = fh)
              if not pargs.hltpath:
                  print('#'+','.join( [ '%d'%nfills,'%d'%nruns,'%d'%nls,'%d'%ncmsls,myformatter.lumi(totdelivered),myformatter.lumi(totrecorded)] ),file=fh)
              else:
                  totalpathdelivered = 0
                  totalpathrecorded = 0
                  for pentry in hltpathSummary:
                      print('#'+','.join( [ '%s'%pentry[0],'%d'%pentry[1],'%d'%pentry[2],'%d'%pentry[3], myformatter.lumi(pentry[4]),myformatter.lumi(pentry[5]) ] ), file=fh)
                      totalpathdelivered =  totalpathdelivered+pentry[4]
                      totalpathrecorded = totalpathrecorded+pentry[5]
                  print('#',file=fh)
                  print('#Sum delivered  : %s'%myformatter.lumi(totalpathdelivered),file=fh)
                  print('#Sum recorded : %s'%myformatter.lumi(totalpathrecorded),file=fh)
          if parsediffruns or parsediffls:
              print ('\nWarning: problems found in merging -i and --normtag selections:')
              if parsediffruns:
                  print ('  runs %s are not covered by normtag'%(parsediffruns))
              if parsediffls:
                  for r in parsediffls.keys():
                      print ('  in run %d %s is not a superset of %s'%(r, parsediffls[r][0], parsediffls[r][1]))
          print         

          if checkjson:
              selectlist = []
              for k,v in rselect.iteritems():
                  for vv in v:
                      rlist = api.expandrange(vv)
                      for r in rlist:
                          selectlist.append((k,r))
              arr_selectlist = np.array(selectlist,dtype=np.dtype('int,int'))
              arr_returnedls = np.array(g_returnedls,dtype=np.dtype('int,int'))
              clist = list(np.setdiff1d(arr_selectlist,arr_returnedls,assume_unique=True))
              if pargs.totable:
                  print ('#Check JSON:')
                  print ('#(run,ls) in json but not in results: %s'%(str(clist)) )
              else:
                  print('#Check JSON:',file=fh)
                  print('#(run,ls) in json but not in results: %s'%(str(clist)),file=fh)

          if g_nulldeadtime:
              if pargs.totable:
                  print ('#WARN: unknown deadtime while CMS is on')
                  for rr,lss in g_nulldeadtime.items():
                      print ( '#%d %s'%(rr,str(lss)) )
              else:
                  print('#WARN: unknown deadtime while CMS is on',file=fh)
                  for rr,lss in g_nulldeadtime.items():
                      print('#%d %s'%(rr,str(lss)),file=fh)
                      
          if fh and fh is not sys.stdout: 
              fh.close()
          
          sys.exit(0)

      elif args['<command>'] == 'beam':
          from . import brilcalc_beam

          parseresult = docopt.docopt(brilcalc_beam.__doc__,argv=cmmdargv)
          parseresult = brilcalc_beam.validate(parseresult)
          ##parse selection params
          pargs = clicommonargs.parser(parseresult)

          ##db params
          dbschema = ''
          if not pargs.dbconnect.find('oracle')!=-1: dbschema = 'cms_lumi_prod'
          if sys.version_info > (3,):
              dbengine = create_engine(pargs.connecturl,max_identifier_length=128)
          else:
              dbengine = create_engine(pargs.connecturl)
          totz=utctmzone
          if pargs.cerntime: totz=cerntmzone
          if pargs.tssec: totz=None
          ##display params          
          fh = None
          ptable = None
          csvwriter = None

          (datatagname,datatagnameid) = findtagname(dbengine,pargs.datatagname,dbschema)
          log.debug('datatagname: %s, datatagnameid: %d'%(datatagname,datatagnameid))    
          header = ['fill','run','ls','time','egev','intensity1','intensity2','ncollidingbx']
          if pargs.withBX:
              header = ['fill','run','ls','time','[bxidx intensity1 intensity2]']
          if not pargs.totable:
              fh = pargs.ofilehandle
              print('#Data tag : %s'%(datatagname), file = fh)
              print('#'+','.join(header), file=fh)
              csvwriter = csv.writer(fh)
          else:
              ptable = display.create_table(header,header=True,maxwidth=80,align='l')                        

          idfields = ['fillnum','runnum','lsnum','timestampsec','beamstatusid','numbxbeamactive']    
          fields = ['egev','intensity1','intensity2']
          if pargs.withBX:
              fields = ['bxidxblob','bxintensity1blob','bxintensity2blob']
              
          rselectrange = []
          if pargs.runlsSeries is not None:
              for r,l in pargs.runlsSeries.iteritems():
                  if not r in rselectrange:
                      rselectrange.append(r)          
          shards = api.locate_shards(dbengine,runmin=pargs.runmin,runmax=pargs.runmax,fillmin=pargs.fillmin,fillmax=pargs.fillmax,tssecmin=pargs.tssecmin,tssecmax=pargs.tssecmax,orrunlist=rselectrange,schemaname=dbschema)
          if not shards:
              print ( 'Failed to find data table for the requested time range.' )
              sys.exit(1)
          log.debug('shards: '+str(shards))
          for shard in shards:
              beamIt = api.dataIter(dbengine,'beam','',shard,datafields=fields,idfields=idfields,schemaname=dbschema,fillmin=pargs.fillmin,fillmax=pargs.fillmax,runmin=pargs.runmin,runmax=pargs.runmax,amodetagid=pargs.amodetagid,targetegev=pargs.egev,beamstatusid=pargs.beamstatusid,tssecmin=pargs.tssecmin,tssecmax=pargs.tssecmax,runlsselect=pargs.runlsSeries,sorted=True,datatagnameid=datatagnameid)
              if not beamIt:
                  continue
              for row in beamIt:
                  fillnum = row['fillnum']
                  runnum = row['runnum']
                  lsnum = row['lsnum']                          
                  timestampsec = row['timestampsec']
                  dtime = str(timestampsec)
                  ncollidingbx = 0
                  if totz is not None:
                      d = datetime.fromtimestamp(int(timestampsec),tz=pytz.utc)
                      dtime = d.astimezone(totz).strftime(params._datetimefm) 
                  if pargs.withBX:
                      bxintensity = None
                      bxintensitystr = '[]'
                      if 'bxidxblob' in row and row['bxidxblob'] is not None:
                          bxidxarray = np.array(api.unpackBlobtoArray(row['bxidxblob'],'H'))                            
                          bxidxarray = bxidxarray[bxidxarray!=np.array(None)]
                          if bxidxarray is not None and bxidxarray.size>0:
                              bxintensity1array =  np.array(api.unpackBlobtoArray(row['bxintensity1blob'],'f'))
                              bxintensity2array =  np.array(api.unpackBlobtoArray(row['bxintensity2blob'],'f'))
                              bxintensity = np.transpose( np.array([bxidxarray+1,bxintensity1array,bxintensity2array]) )
                              a = map(fm.bxintensity,bxintensity)                          
                              bxintensitystr = '['+' '.join(a)+']'
                              del bxintensity1array
                              del bxintensity2array
                          del bxidxarray
                          display.add_row( ['%d'%fillnum,'%d'%runnum,'%d'%lsnum,dtime,'%s'%bxintensitystr], fh=fh, csvwriter=csvwriter, ptable=ptable )
                  else:
                      egev = row['egev']
                      intensity1 = row['intensity1']/pargs.scalefactor
                      intensity2 = row['intensity2']/pargs.scalefactor
                      ncollidingbx = row['numbxbeamactive'] 
                      display.add_row( ['%d'%fillnum,'%d'%runnum,'%d'%lsnum,dtime,'%.1f'%egev,'%.4e'%intensity1,'%.4e'%intensity2, '%d'%ncollidingbx],fh=fh, csvwriter=csvwriter, ptable=ptable)

          if pargs.totable:
              print ('#Data tag : ',datatagname)
              display.show_table(ptable,pargs.outputstyle)
              del ptable
                  
          if fh and fh is not sys.stdout: fh.close()
          sys.exit(0)
          
      elif args['<command>'] == 'trg':      
          from . import brilcalc_trg
          parseresult = docopt.docopt(brilcalc_trg.__doc__,argv=cmmdargv)
          parseresult = brilcalc_trg.validate(parseresult)
          ##parse selection params
          pargs = clicommonargs.parser(parseresult)
          if not pargs.runmin:
              print('-r runnumber option is required for trg command')
              sys.exit(0)
          ##db params
          dbschema = ''
          if not pargs.dbconnect.find('oracle')!=-1: dbschema = 'cms_lumi_prod'
          if sys.version_info > (3,):
              dbengine = create_engine(pargs.connecturl,max_identifier_length=128)
          else:
              dbengine = create_engine(pargs.connecturl)
          ##display params          
          fh = None
          ptable = None
          csvwriter = None

          is_pathinfo = parseresult['--pathinfo']
          is_prescale = parseresult['--prescale']
            
          if is_pathinfo:
              header = ['hltpath','logic','l1bit']
              hltrunconfig = []
              hltconfigids = []
              if pargs.hltconfigid or pargs.hltkey or pargs.runmin:
                  hltrunconfig = api.get_hltrunconfig(dbengine,hltconfigid=pargs.hltconfigid,hltkey=pargs.hltkey,runnum=pargs.runmin,schemaname=dbschema)  #[['runnum','hltconfigid','hltkey'],['runnum','hltconfigid','hltkey']]
                  if not hltrunconfig:
                      print ('hltconfig not found')
                      sys.exit(0)
                  hltconfigids = np.unique( [h[1] for h in hltrunconfig] )        
              hltl1map = api.get_hlttrgl1seedmap(dbengine,hltpath=pargs.hltpath,hltconfigids=hltconfigids,schemaname=dbschema)
              # {hltconfigid: [[hltpathid,hltpathname,seedtype,[seedvalues]]]}
              if not hltl1map:
                  print ('hltl1seed mapping not found')
                  sys.exit(0)              
              if pargs.dataset is not None:
                    hltconfigid = int(hltconfigids[0])
                    hltpathid = int(hltl1map[hltconfigid][0][0])

                    r = api.is_hltpathid_in_dataset(dbengine,hltpathid,pargs.dataset,hltconfigid,schemaname=dbschema)
                    ##check hltpathid and dataset are in the same menu
                    if not r:
                      print( 'dataset {} and hltpath{} are not in the same menu'.format(pargs.dataset,pargs.hltpath) )
                      sys.exit(0)
                      
              if not pargs.totable:
                  fh = pargs.ofilehandle
                  print('# '+','.join(header),file=fh)
                  csvwriter = csv.writer(fh)
              else:
                  ptable = display.create_table(header,header=True,maxwidth=60,align='l')
                                
              for hid in hltl1map.keys():
                  hdata = hltl1map[hid]
                  for pathdata in hdata:
                      hltpathname = pathdata[1]
                      seedtype = pathdata[2]
                      seedvals = pathdata[3]
                      seedstr = ' '.join([str(i) for i in seedvals])
                      display.add_row( [ '%s'%hltpathname , '%s'%seedtype, '%s'%seedstr], fh=fh, csvwriter=csvwriter, ptable=ptable )                
              if ptable:
                  display.show_table(ptable,pargs.outputstyle)         
              del ptable
          elif is_prescale:
              header = ['run','cmsls','prescidx']
              if pargs.hltpath is not None:
                  if pargs.dataset is not None:
                    header = header+['totprescval','dataset prescval','hltpath/prescval','logic','l1bit/prescval']
                  else:
                    header = header+['totprescval','hltpath/prescval','logic','l1bit/prescval']
                    
              hltrunconfig = api.get_hltrunconfig(dbengine,hltconfigid=pargs.hltconfigid,hltkey=pargs.hltkey,runnum=pargs.runmin,schemaname=dbschema)  #[['runnum','hltconfigid','hltkey'],['runnum','hltconfigid','hltkey']]
              runlist = [x[0] for x in hltrunconfig]
              hltconfigids = [x[1] for x in hltrunconfig]
              presc = api.get_prescidx_change(dbengine,runlist,schemaname=dbschema)
              #{runnum:[[lslastscaler,prescidx]]
                           
              if not pargs.totable:
                  fh = pargs.ofilehandle
                  print('# '+','.join(header),file=fh)
                  csvwriter = csv.writer(fh)
              else:
                  ptable = display.create_table(header,header=True,maxwidth=60,align='l')
                  
              if pargs.hltpath:
                  hltl1map = api.get_hlttrgl1seedmap(dbengine,hltpath=pargs.hltpath,hltconfigids=hltconfigids,schemaname=dbschema)
                  # {hltconfigid: [[hltpathid,hltpathname,seedtype,[seedvalues]]]}                   
                  if not hltl1map:
                      print ('No hltpathl1seed mapping found')
                      sys.exit(0)
                      
                  selected_hltconfigids = hltl1map.keys()
                  if not selected_hltconfigids:
                      print ('No hltconfig found ')
                      sys.exit(0)
                      
                  datasetpresc = 1
                  if pargs.dataset is not None:
                    hltconfigid = list(selected_hltconfigids)[0]
                    hltpathid = hltl1map[hltconfigid][0][0]
                    r = api.is_hltpathid_in_dataset(dbengine,hltpathid,pargs.dataset,hltconfigid,schemaname=dbschema)
                    ##check hltpathid and dataset are in the same menu
                    if not r:
                      print( 'dataset {} and hltpath {} are not in the same menu'.format(pargs.dataset,pargs.hltpath) )
                      sys.exit(0)

                    datasetprescs = api.get_dataset_presc(dbengine,hltconfigid,pargs.dataset,schemaname=dbschema)#{(lsnum,prescidx):prescval}
                    
                  for [runnum,hltconfigid,hltkey] in [r for r in hltrunconfig if r[1] in selected_hltconfigids]:
                      if hltconfigid not in hltl1map:
                          continue
                      mdata = hltl1map[hltconfigid]
                      for [hltpathid,hltpathname,l1seedlogic,l1c] in mdata:                          
                          if runnum not in presc:
                              display.add_row( [ '%d'%runnum, '%s'%'None', '%s'%'None', '%d'%1,'%s'%hltpathname, '%s'%l1seedlogic, '%s'%' '.join(l1c)], fh=fh, csvwriter=csvwriter, ptable=ptable )
                          else:
                              hltpathnamemap = dict( [(h[0],[h[1],h[2]]) for h in mdata] )
                              if hltpathid not in hltpathnamemap:
                                  continue
                              for [lsnum,prescidx] in presc[runnum]:
                                  hltprescval = api.get_hltprescale(dbengine,runnum,lsnum,hltconfigid,prescidx,hltpathid,schemaname=dbschema)
                                  rl = api.get_l1prescale(dbengine,runnum,lsnum,l1candidates=l1c,prescidxs=prescidx,ignorel1mask=parseresult['--ignore-mask'] ,schemaname=dbschema)
                                  #{(prescidx,l1bitname):[l1prescval,bitmask]}
                                  if not rl: continue
                                  (hltpathname,l1seedlogic) = hltpathnamemap[hltpathid]                                  
                                  l1keys = [ k for k in rl.keys() if k[0]==prescidx ]
                                  l1bitnames = [ v[1] for v in l1keys ]
                                  l1vals = [ rl[k] for k in l1keys ]                       
                                  l1prescvals = [ v[0] for v in l1vals ]
                                  l1bits = zip(l1bitnames,l1prescvals)
                                  l1inner = map(fm.bitprescFormatter,l1bits)
                                  l1bitsStr = ' '.join(l1inner)
                                  hltpathStr = '/'.join([hltpathname,str(hltprescval)])
                                  totpresc = totalprescaleNEW(hltprescval,l1seedlogic,l1prescvals)
                                  if pargs.dataset is not None:
                                      datasetpresc = datasetprescs[(lsnum,prescidx)]
                                      #print('datasetpresc val ',lsnum,prescidx,datasetpresc)
                                      totpresc = totpresc*datasetpresc
                                      display.add_row( [ '%d'%runnum, '%d'%lsnum,'%d'%prescidx, '%.2f'%datasetpresc ,'%.2f'%totpresc,'%s'%hltpathStr, '%s'%l1seedlogic, '%s'%l1bitsStr], fh=fh, csvwriter=csvwriter, ptable=ptable )
                                  else:
                                      display.add_row( [ '%d'%runnum, '%d'%lsnum, '%d'%prescidx, '%.2f'%totpresc,'%s'%hltpathStr, '%s'%l1seedlogic, '%s'%l1bitsStr], fh=fh, csvwriter=csvwriter, ptable=ptable ) 
                  del hltl1map
              else:
                  if not presc:
                      print ('Not prescaled')
                      sys.exit(0)
                  for runnum in sorted(presc.keys()):
                      for plsdata in presc[runnum]:
                          lsnum = plsdata[0]
                          prescidx = plsdata[1]
                          display.add_row( [ '%d'%runnum, '%d'%lsnum, '%d'%prescidx ], fh=fh, csvwriter=csvwriter, ptable=ptable ) 
              if ptable:
                  display.show_table(ptable,pargs.outputstyle)         
                  del ptable          
          else:
              hltrunconfig = api.get_hltrunconfig(dbengine,hltconfigid=pargs.hltconfigid,hltkey=pargs.hltkey,runnum=pargs.runmin,schemaname=dbschema)
              if not hltrunconfig:
                  print ('hltconfig not found')
                  sys.exit(0)
              header = ['hltconfigid','hltkey','run']    
              if not pargs.totable:
                  fh = pargs.ofilehandle
                  print('# '+','.join(header),file=fh)
                  csvwriter = csv.writer(fh)
              else:
                  ptable = display.create_table(header,header=True,maxwidth=80,align='l')

              hltrunconfig_df = pd.DataFrame(hltrunconfig,columns=['runnum','hltconfigid','hltkey'])
              grouped = hltrunconfig_df.groupby(['hltconfigid','hltkey'])              
              for name,group in grouped:
                  hltconfigid = int(name[0])
                  hltkey = str(name[1])
                  runsStr = ','.join([str(i) for i in group['runnum'].values])
                  display.add_row( [ '%d'%hltconfigid, '%s'%hltkey , '%s'%runsStr], fh=fh, csvwriter=csvwriter, ptable=ptable )                
              if ptable:
                  display.show_table(ptable,pargs.outputstyle)         
                  del ptable                  
              del hltrunconfig_df
          if fh and fh is not sys.stdout: fh.close()
          sys.exit(0)
          
    except docopt.DocoptExit:
      raise docopt.DocoptExit('Error: incorrect input format for '+args['<command>'])            
    except schema.SchemaError as e:
      exit(e)    
    return

if __name__ == '__main__':
    brilcalc_main()
    exit(0)



"""
Usage:
  brilcalc lumi [options] 

Options:
  -h, --help                    Show this screen
  -c CONNECT                    Service name [default: offline]
  -p AUTHPATH                   Authentication file
  -n SCALEFACTOR                Scale factor to results [default: 1.0]
  -f FILLNUM                    Fill number
  -r RUNNUMBER                  Run number
  -i INPUTFILE                  Input selection json file or string
  -o OUTPUTFILE                 Output csv file. Special file '-' for stdout.
  -b BEAMSTATUS                 Beam mode. FLAT TOP,SQUEEZE,ADJUST,STABLE BEAMS
  -u UNIT                       Lumi unit. hz/ub,1e30/cm2s,/nb,1e33/cm2 [default: /ub]
  --filedata FILEDATA           Lumi data hdf5 file or directory 
  --precision PRECISION         Use result precision [default: 9f]
  --amodetag AMODETAG           Accelerator mode 
  --beamenergy BEAMENERGY       Target single beam energy in GeV
  --minBiasXsec MINBIASXSEC     Minbias cross-section in ub
  --datatag DATATAG             Data tag name
  --normtag NORMTAG             correction tag or combined correction tag selection file or string
  --begin BEGIN                 Min start time (MM/DD/YY HH24:MI:SS)/fill/run 
  --end END                     Max start time (MM/DD/YY HH24:MI:SS)/fill/run
  --output-style OSTYLE         Screen output style. tab, html, csv [default: tab]
  --type LUMITYPE               Luminosity type. hfoc,bcm1f,bcm1fsi,bcm1futca,pltzero,pxl,dt,hfet,radmmon,ramses
  --hltpath HLTPATH             HLT path name or fnmatch pattern
  --dataset DATASET             dataset name
  --byls                        Show result in ls granularity
  --xing                        Show result in bx granularity
  --without-correction          Show raw data without calibration
  --ignore-mask                 Ignore l1 trgger bit mask for --hltpath specific luminosity
  --cerntime                    Show time in CERN local time
  --tssec                       Show time as second since Epoch
  --xingId BXID                 Filter bunches by id, "1,2,3" or file
  --xingTr BXTHRESHOLD          Filter bunches by percentage of max lumi, 0-1 
  --xingMin XINGMIN             Filter bunches by min lumi value   
  --without-checkjson           Do not check total returned lumi sections against input json file.  

"""

import os,sys
from docopt import docopt
from schema import Schema
from brilws.cli import clicommonargs

def validate(optdict):
    result={}
    #extract sub argdict here
    myvalidables = ['-c','-n','-f','-r','-i','-o','--amodetag','-b','--beamenergy','--minBiasXsec','--datatag','--normtag','--begin','--end','--output-style','--type','--hltpath','--xingId','--xingTr','--xingMin','--precision','--filedata',str]
    argdict = dict((k,v) for k,v in clicommonargs.argvalidators.items() if k in myvalidables)
    
    schema = Schema(argdict)
    result = schema.validate(optdict)
    if not result['-i'] and not result['-f'] and not result['-r'] and not result['--begin']:
        print ('Error: at least one time selection option in %s is required'%(','.join(['-i','-f','-r','--begin'])))
        sys.exit(0)
    if result['--filedata'] and not result['--byls']:
        print ('Error: --filedata can only be used with --byls')
        sys.exit(0)            
    return result    
if __name__ == '__main__':
    args = docopt(__doc__,options_first=True)
    print (args)




#import pandas as pd
import numpy as np
import prettytable

_floatformatter='{:,.3f}'.format

#pd.set_option('display.max_columns', 4)
#pd.set_option('display.max_colwidth', 200)
#pd.set_option('display.max_rows', 1000)
#pd.set_option('display.width', 200)
#pd.set_option('display.float_format',_floatformatter)



_prettytable_style={
    'withheader':True,
    'align':'l',
    'maxwidth':80
}    


def add_row(row,fh=None,csvwriter=None,ptable=None):
    '''
    input: row=[]
    '''
    if fh and csvwriter:
        csvwriter.writerow(row)
    elif ptable:
        ptable.add_row(row)
    else:
        return
    
def show_table(ptable,style='tab'):
    if not ptable: return
    if style=='tab':
        print(ptable.get_string())
    elif style=='html':
        print(ptable.get_html_string())
    
def create_table(columns,**kwargs):
    '''
    create a prettytable instance
    https://code.google.com/p/prettytable/wiki/Tutorial
    '''
    ptable = prettytable.PrettyTable(columns)
    if 'border' in kwargs: ptable.border = kwargs['border']
    if 'header' in kwargs: ptable.header = kwargs['header']
    if 'header_style' in kwargs: ptable.header_style = kwargs['header_style']
    if 'hrules' in kwargs: ptable.hrules = kwargs['hrules']
    if 'vrules' in kwargs: ptable.hrules = kwargs['vrules']
    if 'align' in kwargs: ptable.align = kwargs['align']
    if 'valign' in kwargs: ptable.valign = kwargs['valign']
    if 'int_format' in kwargs: ptable.int_format = kwargs['int_format']
    if 'float_format' in kwargs: ptable.float_format = kwargs['float_format']
    if 'padding_width' in kwargs: ptable.padding_width = kwargs['padding_width']
    if 'left_padding_width' in kwargs: ptable.left_padding_width = kwargs['left_padding_width']
    if 'right_padding_width' in kwargs: ptable.right_padding_width = kwargs['right_padding_width']
    if 'vertical_char' in kwargs: ptable.vertical_char = kwargs['vertical_char']
    if 'horizontal_char' in kwargs: ptable.horizontal_char = kwargs['horizontal_char']
    if 'junction_char' in kwargs: ptable.junction_char = kwargs['junction_char']
    if 'maxwidth' in kwargs: ptable.max_width = kwargs['maxwidth'] 
    return ptable


def formatter_tuple( x_y ):
    x,y = x_y
    if isinstance(x,float):
        x = _floatformatter(x)
    if isinstance(y,float):
        y = _floatformatter(y)
    return '%s:%s'%(x,y)

def listdf(df,npp=100, columns=None, formatters=None, index=False, justify='left', pagination=False):
    '''
    Inputs:
        df:         dataframe
        npp:        number of records per page (default=50)
        formatters: field formatter (default=None)
        index:      display row id (default=False)
    '''
    if not pagination:
        print (df.to_string(columns=columns,formatters=formatters,index=index, justify=justify))
    return
    nrows, ncols = df.shape
    total_pages = nrows/npp + 1 
    rec_last_pg = nrows % npp # number of records in last page
    ptr = 0
    for i in xrange(total_pages):
        fromrow = ptr
        if i==(total_pages-1):
           torow = fromrow + rec_last_pg -1
        else:
           torow = fromrow + npp -1
        ptr = torow +1
        print (df.ix[fromrow:torow,:].to_string(columns=columns,formatters=formatters,index=index, justify=justify))
        if i!=(total_pages-1):
            try: 
                raw_input("Press a key to continue or '^C' to break...")
            except KeyboardInterrupt:
                break

if __name__=='__main__':
    chunksize = 200
    #rowdef = np.dtype([('tagid',np.uint64),('tagname',object),('value',np.float32)])
    rowdef = np.dtype([('tagid','u8'),('tagname','O'),('value','f4')]) 
    mydf = np.empty( (chunksize,), dtype=rowdef )
    myformatter = {'value':_floatformatter}
    for i in xrange(chunksize):
        mydf['tagname'][i]='ab'
        mydf['tagid'][i]=i
        mydf['value'][i]=0.5*i
    #df = pd.DataFrame.from_records(mydf,index=None,columns=rowdef.names)
    #print rowdef.names
    #listdf(df,npp=30,formatters=myformatter,columns=['tagid','tagname'])

    
    create_table(['a','b'],border = True,align='l')

       



"""
Usage:
  brilschema loadmap [options] 

Options:
  -h, --help       Show this screen
  --name NAME      Map: datatable,amodetag,beamstatus,trgbit,hltconfig,l1seed,hltpath,trghltseed,dataset,hltstreamdataset 
  -i INPUT         Input csv file or database
  -o OUTPUT        Output csv file or Database 
  -p AUTH          Path to authentication.ini 

"""


import os,re
from docopt import docopt
from schema import Schema, And, Or, Use
choices_mapdata = ['datatable','amodetag','beamstatus','trgbit','hltconfig','l1seed','hltpath','trghltseed','dataset','hltstreamdataset']


def validate(optdict):
    result={}
    schema = Schema({ 
     '-i': And(str,lambda s: os.path.isfile(os.path.expanduser(s)) or s.find('sqlite:')!=-1 or s.find('oracle:')!=-1 ,error='-i SOURCE should be a csv file or database connect string'),
     '-o': And(str,lambda s: os.path.isfile(os.path.expanduser(s)) or s.find('sqlite:')!=-1 or s.find('oracle:')!=-1 ,error='-o DEST should be a csv file or database connect string'),     
     '--name': And(str,lambda s: s.lower() in choices_mapdata, error='--name must be in '+str(choices_mapdata)),
     '-p': Or(None,And(str,lambda s: os.path.isfile(os.path.join(os.path.expanduser(s),'authentication.ini'))) ,error='-p AUTH should contain authentication.ini'),   
     str:object # catch all
    })
    result=schema.validate(optdict)    
    return result

if __name__ == '__main__':
    print (docopt(__doc__,options_first=True))
